from pystage.core import CoreStage
from pystage.en import Sprite


class Stage():

    def __init__(self):
        self._core = CoreStage()
        self._core.facade = self
        self._core.sprite_facade_class = Sprite

    def add_a_sprite(self, costume="default") -> Sprite:
        return self._core.pystage_createsprite(costume=costume)

    def play(self):
        self._core.pystage_play()

    def create_clone_of_sprite(self, sprite='_myself_'):
        """create clone of %1

        Translation string: create clone of %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.control_create_clone_of_sprite(sprite)

    def stop_all(self):
        """stop all

        Translation string: stop all
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.control_stop_all()

    def stop_other_scripts_in_sprite(self):
        """stop other scripts in sprite

        Translation string: stop other scripts in sprite
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.control_stop_other()

    def stop_this_script(self):
        """stop this script

        Translation string: stop this script
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.control_stop_this()

    def wait_seconds(self, secs):
        """wait %1 seconds

        Translation string: wait %1 seconds
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        secs : FILL
        

        Returns
        -------

        """
        return self._core.control_wait(secs)

    def change_by(self, name, value):
        """change %1 by %2

        Translation string: change %1 by %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        value : FILL
        

        Returns
        -------

        """
        return self._core.data_changevariableby(name, value)

    def hide_variable(self, name):
        """hide variable %1

        Translation string: hide variable %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.data_hidevariable(name)

    def set_variable(self, name, value):
        """set %1 to %2

        Translation string: set %1 to %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        value : FILL
        

        Returns
        -------

        """
        return self._core.data_setvariableto(name, value)

    def show_variable(self, name):
        """show variable %1

        Translation string: show variable %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.data_showvariable(name)

    def get_variable(self, name):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.data_variable(name)

    def broadcast(self, message):
        """broadcast %1

        Translation string: broadcast %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        message : FILL
        

        Returns
        -------

        """
        return self._core.event_broadcast(message)

    def broadcast_and_wait(self, message):
        """broadcast %1 and wait

        Translation string: broadcast %1 and wait
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        message : FILL
        

        Returns
        -------

        """
        return self._core.event_broadcastandwait(message)

    def when_backdrop_switches_to(self, backdrop, generator_function, name='', no_refresh=False):
        """when backdrop switches to %1

        Translation string: when backdrop switches to %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        backdrop : FILL
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.event_whenbackdropswitchesto(backdrop, generator_function, name, no_refresh)

    def when_i_receive(self, message, generator_function, name='', no_refresh=False):
        """when I receive %1

        Translation string: when I receive %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        message : FILL
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.event_whenbroadcastreceived(message, generator_function, name, no_refresh)

    def when_GREENFLAG_clicked(self, generator_function, name='', no_refresh=False):
        """when <greenflag> clicked

        Translation string: when <greenflag> clicked
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.event_whenflagclicked(generator_function, name, no_refresh)

    def when_loudness_GREATERTHAN(self, value, generator_function, name='', no_refresh=False):
        """when loudness <greater> %2

        Translation string: when loudness <greater> %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.event_whengreaterthan_loudness(value, generator_function, name, no_refresh)

    def when_timer_GREATERTHAN(self, value, generator_function, name='', no_refresh=False):
        """when timer <greater> %2

        Translation string: when timer <greater> %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.event_whengreaterthan_timer(value, generator_function, name, no_refresh)

    def when_key_pressed(self, key, generator_function, name='', no_refresh=False):
        """when %1 key pressed

        Translation string: when %1 key pressed
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        key : FILL
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.event_whenkeypressed(key, generator_function, name, no_refresh)

    def when_this_sprite_clicked(self, generator_function, name='', no_refresh=False):
        """when this sprite clicked

        Translation string: when this sprite clicked
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.event_whenthisspriteclicked(generator_function, name, no_refresh)

    def backdrop_name(self):
        """backdrop name

        Translation string: backdrop name
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_backdropnumbername_name()

    def backdrop_number(self):
        """backdrop number

        Translation string: backdrop number
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_backdropnumbername_number()

    def change_brightness_effect_by(self, value):
        """change brightness effect by %2

        Translation string: change brightness effect by %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_changeeffectby_brightness(value)

    def change_color_effect_by(self, value):
        """change color effect by %2

        Translation string: change color effect by %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_changeeffectby_color(value)

    def change_fisheye_effect_by(self, value):
        """change fisheye effect by %2

        Translation string: change fisheye effect by %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_changeeffectby_fisheye(value)

    def change_ghost_effect_by(self, value):
        """change ghost effect by %2

        Translation string: change ghost effect by %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_changeeffectby_ghost(value)

    def change_mosaic_effect_by(self, value):
        """change mosaic effect by %2

        Translation string: change mosaic effect by %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_changeeffectby_mosaic(value)

    def change_pixelate_effect_by(self, value):
        """change pixelate effect by %2

        Translation string: change pixelate effect by %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_changeeffectby_pixelate(value)

    def change_whirl_effect_by(self, value):
        """change whirl effect by %2

        Translation string: change whirl effect by %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_changeeffectby_whirl(value)

    def clear_graphic_effects(self):
        """clear graphic effects

        Translation string: clear graphic effects
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_cleargraphiceffects()

    def next_backdrop(self):
        """next backdrop

        Translation string: next backdrop
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_nextbackdrop()

    def set_brightness_effect_to(self, value):
        """set brightness effect to %2

        Translation string: set brightness effect to %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_seteffectto_brightness(value)

    def set_color_effect_to(self, value):
        """set color effect to %2

        Translation string: set color effect to %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_seteffectto_color(value)

    def set_fisheye_effect_to(self, value):
        """set fisheye effect to %2

        Translation string: set fisheye effect to %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_seteffectto_fisheye(value)

    def set_ghost_effect_to(self, value):
        """set ghost effect to %2

        Translation string: set ghost effect to %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_seteffectto_ghost(value)

    def set_mosaic_effect_to(self, value):
        """set mosaic effect to %2

        Translation string: set mosaic effect to %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_seteffectto_mosaic(value)

    def set_pixelate_effect_to(self, value):
        """set pixelate effect to %2

        Translation string: set pixelate effect to %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_seteffectto_pixelate(value)

    def set_whirl_effect_to(self, value):
        """set whirl effect to %2

        Translation string: set whirl effect to %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_seteffectto_whirl(value)

    def switch_backdrop_to(self, backdrop):
        """switch backdrop to %1

        Translation string: switch backdrop to %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        backdrop : FILL
        

        Returns
        -------

        """
        return self._core.looks_switchbackdropto(backdrop)

    def switch_backdrop_to_and_wait(self, backdrop):
        """switch backdrop to %1 and wait

        Translation string: switch backdrop to %1 and wait
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        backdrop : FILL
        

        Returns
        -------

        """
        return self._core.looks_switchbackdroptoandwait(backdrop)

    def of(self, operator, number):
        """%1 of %2

        Translation string: %1 of %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        operator : FILL
        number : FILL
        

        Returns
        -------

        """
        return self._core.operator_mathop(operator, number)

    def pick_random_to(self, start, end):
        """pick random %1 to %2

        Translation string: pick random %1 to %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        start : FILL
        end : FILL
        

        Returns
        -------

        """
        return self._core.operator_random(start, end)

    def add_backdrop(self, name, center_x=None, center_y=None):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        center_x : FILL
        center_y : FILL
        

        Returns
        -------

        """
        return self._core.pystage_addbackdrop(name, center_x, center_y)

    def pystage_addsound(self, name):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.pystage_addsound(name)

    def pystage_createsprite(self, costume='default'):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        costume : FILL
        

        Returns
        -------

        """
        return self._core.pystage_createsprite(costume)

    def pystage_insertbackdrop(self, index, name, center_x=None, center_y=None):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        index : FILL
        name : FILL
        center_x : FILL
        center_y : FILL
        

        Returns
        -------

        """
        return self._core.pystage_insertbackdrop(index, name, center_x, center_y)

    def create_variable(self, name, all_sprites=True):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        all_sprites : FILL
        

        Returns
        -------

        """
        return self._core.pystage_makevariable(name, all_sprites)

    def pystage_play(self):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.pystage_play()

    def pystage_replacebackdrop(self, index, name, center_x=None, center_y=None):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        index : FILL
        name : FILL
        center_x : FILL
        center_y : FILL
        

        Returns
        -------

        """
        return self._core.pystage_replacebackdrop(index, name, center_x, center_y)

    def set_monitor_position(self, name, x, y):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        x : FILL
        y : FILL
        

        Returns
        -------

        """
        return self._core.pystage_setmonitorposition(name, x, y)

    def pystage_setmonitorstyle_large(self, name):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.pystage_setmonitorstyle_large(name)

    def pystage_setmonitorstyle_normal(self, name):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.pystage_setmonitorstyle_normal(name)

    def pystage_setmonitorstyle_slider(self, name):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.pystage_setmonitorstyle_slider(name)

    def answer(self):
        """answer

        Translation string: answer
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_answer()

    def ask_and_wait(self, question):
        """ask %1 and wait

        Translation string: ask %1 and wait
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        question : FILL
        

        Returns
        -------

        """
        return self._core.sensing_askandwait(question)

    def current_date(self):
        """current date

        Translation string: current date
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_current_date()

    def current_day_of_week(self):
        """current day of week

        Translation string: current day of week
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_current_dayofweek()

    def current_hour(self):
        """current hour

        Translation string: current hour
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_current_hour()

    def current_minute(self):
        """current minute

        Translation string: current minute
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_current_minute()

    def current_month(self):
        """current month

        Translation string: current month
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_current_month()

    def current_second(self):
        """current second

        Translation string: current second
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_current_second()

    def current_year(self):
        """current year

        Translation string: current year
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_current_year()

    def days_since(self):
        """days since 2000

        Translation string: days since 2000
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_dayssince2000()

    def key_pressed(self, key):
        """key %1 pressed?

        Translation string: key %1 pressed?
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        key : FILL
        

        Returns
        -------

        """
        return self._core.sensing_keypressed(key)

    def loudness(self):
        """loudness

        Translation string: loudness
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_loudness()

    def mouse_down(self):
        """mouse down?

        Translation string: mouse down?
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_mousedown()

    def mouse_x(self):
        """mouse x

        Translation string: mouse x
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_mousex()

    def mouse_y(self):
        """mouse y

        Translation string: mouse y
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_mousey()

    def backdrop_name_of(self, stage='_stage_'):
        """backdrop name of %2

        Translation string: backdrop name of %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        stage : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_backdropname(stage)

    def backdrop_of(self, stage='_stage_'):
        """backdrop # of %2

        Translation string: backdrop # of %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        stage : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_backdropnumber(stage)

    def costume_name_of(self, sprite):
        """costume name of %2

        Translation string: costume name of %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_costumename(sprite)

    def costume_of(self, sprite):
        """costume # of %2

        Translation string: costume # of %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_costumenumber(sprite)

    def direction_of(self, sprite):
        """direction of %2

        Translation string: direction of %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_direction(sprite)

    def size_of(self, sprite):
        """size of %2

        Translation string: size of %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_size(sprite)

    def of(self, variable, sprite='_stage_'):
        """%1 of %2

        Translation string: %1 of %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        variable : FILL
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_variable(variable, sprite)

    def volume_of(self, sprite='_stage_'):
        """volume of %2

        Translation string: volume of %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_volume(sprite)

    def x_position_of(self, sprite):
        """x position of %2

        Translation string: x position of %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_xposition(sprite)

    def y_position_of(self, sprite):
        """y position of %2

        Translation string: y position of %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_yposition(sprite)

    def reset_timer(self):
        """reset timer

        Translation string: reset timer
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_resettimer()

    def set_drag_mode_draggable(self):
        """set drag mode draggable

        Translation string: set drag mode draggable
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_setdragmode_draggable()

    def set_drag_mode_not_draggable(self):
        """set drag mode not draggable

        Translation string: set drag mode not draggable
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_setdragmode_notdraggable()

    def timer(self):
        """timer

        Translation string: timer
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_timer()

    def username(self):
        """username

        Translation string: username
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_username()

    def change_pan_left_right_effect_by(self, value):
        """change pan left/right effect by %2

        Translation string: change pan left/right effect by %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.sound_changeeffectby_pan(value)

    def change_pitch_effect_by(self, value):
        """change pitch effect by %2

        Translation string: change pitch effect by %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.sound_changeeffectby_pitch(value)

    def change_volume_by(self, value):
        """change volume by %1

        Translation string: change volume by %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.sound_changevolumeby(value)

    def clear_sound_effects(self):
        """clear sound effects

        Translation string: clear sound effects
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sound_cleareffects()

    def start_sound(self, name, loop=0):
        """start sound %1

        Translation string: start sound %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        loop : FILL
        

        Returns
        -------

        """
        return self._core.sound_play(name, loop)

    def play_sound_until_done(self, name):
        """play sound %1 until done

        Translation string: play sound %1 until done
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.sound_playuntildone(name)

    def set_pan_left_right_effect_to(self, value):
        """set pan left/right effect to %2

        Translation string: set pan left/right effect to %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.sound_seteffectto_pan(value)

    def set_pitch_effect_to(self, value):
        """set pitch effect to %2

        Translation string: set pitch effect to %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.sound_seteffectto_pitch(value)

    def set_volume_to(self, value):
        """set volume to %1%

        Translation string: set volume to %1%
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.sound_setvolumeto(value)

    def stop_all_sounds(self):
        """stop all sounds

        Translation string: stop all sounds
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sound_stopallsounds()

    def volume(self):
        """volume

        Translation string: volume
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sound_volume()



import pygame
import pkg_resources


class Monitor(pygame.sprite.Sprite):

    BOLD_FONT_9 = pygame.font.Font(pkg_resources.resource_filename("pystage", "fonts/roboto-bold.ttf"), 9)
    LIGHT_FONT_9 = pygame.font.Font(pkg_resources.resource_filename("pystage", "fonts/roboto-light.ttf"), 9)
    LIGHT_FONT_11 = pygame.font.Font(pkg_resources.resource_filename("pystage", "fonts/roboto-light.ttf"), 11)

    VARIABLE_COLOR = (255, 140, 26)
    MOTION_COLOR = (76, 151, 255)
    SOUND_COLOR = (207, 99, 207)
    SENSING_COLOR = (92, 177, 214)
    BACKGROUND_COLOR = (230, 240, 255)
    BORDER_COLOR = (196, 204, 217)
    TITLE_FONT_COLOR = (87, 94, 117)
    VALUE_FONT_COLOR = (255, 255, 255)

    NORMAL = 1
    LARGE = 2
    SLIDER = 3

    def __init__(self, sprite_or_stage, title, color=VARIABLE_COLOR, mode=1):
        super().__init__()
        self.sprite_or_stage = sprite_or_stage
        # Visible sprites go into this group
        self.sprite_or_stage.stage.monitor_group.add(self)
        self._title = title
        self._value = 0
        self._color = color
        # Pygame coordinates, upper left corner
        self.image = None
        self.rect = None

        # if func is set, it is used to automatically update the monitor
        self._func = None
        self._style = Monitor.NORMAL

        self.update_image()


    def update(self):
        if not self._func:
            return
        # set_value checks if the value is changed
        self.set_value(self._func())


    def update_image(self):
        pos = self.rect.topleft if self.rect else (0,0)

        if self._style == Monitor.NORMAL:
            title_surface = Monitor.BOLD_FONT_9.render(self._title, True, Monitor.TITLE_FONT_COLOR)
            value_surface = Monitor.LIGHT_FONT_9.render(str(self._value), True, Monitor.VALUE_FONT_COLOR)
            
            padding = 10

            rect_w = max(40, value_surface.get_width() + padding)
            rect_h = value_surface.get_height() + padding/2
            rect_surface = pygame.Surface((rect_w, rect_h), pygame.SRCALPHA)
            pygame.draw.rect(rect_surface, self._color, (0,0, rect_w, rect_h), border_radius=3)
            rect_surface.blit(value_surface, (rect_surface.get_width()/2 - value_surface.get_width()/2, padding/4))
            
            w = padding + title_surface.get_width() + padding + rect_surface.get_width() + padding
            h = rect_surface.get_height() + padding/2
            surface = pygame.Surface((w, h), flags=pygame.SRCALPHA)
            pygame.draw.rect(surface, Monitor.BACKGROUND_COLOR, (0,0,w,h), border_radius=3)
            pygame.draw.rect(surface, Monitor.BORDER_COLOR, (0,0,w,h), width=1, border_radius=3)
            surface.blit(title_surface, (padding, padding/2))
            surface.blit(rect_surface, (padding + title_surface.get_width() + padding, padding/4))
        elif self._style == Monitor.LARGE:
            value_surface = Monitor.LIGHT_FONT_11.render(str(self._value), True, Monitor.VALUE_FONT_COLOR)
            
            padding = 8

            rect_w = value_surface.get_width() + padding * 2
            rect_h = value_surface.get_height() + padding * 2
            surface = pygame.Surface((rect_w, rect_h), pygame.SRCALPHA)
            pygame.draw.rect(surface, self._color, (0,0, rect_w, rect_h), border_radius=3)
            pygame.draw.rect(surface, Monitor.BORDER_COLOR, (0,0,rect_w, rect_h), width=1, border_radius=3)
            surface.blit(value_surface, (surface.get_width()/2 - value_surface.get_width()/2, padding))
            
        elif self._style == Monitor.SLIDER:
            # Not yet implemented
            surface = pygame.Surface((10,10), pygame.SRCALPHA)

        self.image = surface
        self.rect = surface.get_rect()
        self.rect.topleft = pos


    def set_position(self, x,y):
        if not self.rect:
            return
        self.rect.x = x + self.sprite_or_stage.stage.center_x
        self.rect.y = -y + self.sprite_or_stage.stage.center_y


    def set_style(self, style):
        if style != self._style:
            self._style = style
            self.update_image()


    def set_value(self, value):
        '''Only change the value via this method!

        The method takes care of the rerendering, if needed.
        '''
        if type(value)==float:
            value = round(value, 2)
            if value == round(value):
                value = round(value)
        if self._value == value:
            return
        self._value = value
        self.update_image()


    def set_function(self, func):
        '''When a function is set, the monitor automatically updates.

        See update(self, dt).
        '''
        self._func = func


    def show(self):
        self.sprite_or_stage.stage.monitor_group.add(self)
        self.update_image()


    def hide(self):
        self.sprite_or_stage.stage.monitor_group.remove(self)

    
    def remove(self):
        del self.sprite_or_stage.monitors[self._title]
        self.kill()



import ast
import inspect

from pystage.l10n.api import get_core_function_from_instance

# Functions that need to be yielded for screen refresh
yield_funcs = [
    "control_wait",
    # Stop needs to yield so that the next command is not executed
    "control_stop_all",
    "control_stop_this",
    "control_stop_other",
    "sound_playuntildone",
    "motion_glidesecstoxy",
    "motion_glideto_random",
    "motion_glideto_sprite",
    "motion_glideto_pointer",
    "sensing_askandwait",
     "looks_sayforsecs",
    "looks_thinkforsecs",
]


class CodeManager():
    def __init__(self, owner):
        # name: code_block
        self.code_blocks: dict[str, CodeBlock] = {}
        # pygame.K_?: [name, ...]
        self.key_pressed_blocks = {}
        # message: [name, ...]
        self.broadcast_blocks = {}
        self.clicked_blocks = []
        self.cloned_blocks = []
        # Name of the code block currently executed.
        # This way, state about the current execustion
        # can be stored safely where it belongs
        self.current_block: CodeBlock = None
        self.owner = owner

    def stop_running_blocks(self):
        for name in self.code_blocks:
            self.code_blocks[name].stop()

    def process_key_pressed(self, key):
        # key is a pygame constant, e.g. pygame.K_a
        # This hat block is special as it only fires again when the code block has ended. 
        # All other hat block methods stop the current execution and restart the block.
        if key in self.key_pressed_blocks:
            for name in self.key_pressed_blocks[key]:
                self.code_blocks[name].start_if_not_running()

    def process_click(self):
        for block in self.clicked_blocks:
            block.start_or_restart()

    def process_broadcast(self, message):
        if message in self.broadcast_blocks:
            for name in self.broadcast_blocks[message]:
                self.code_blocks[name].start_or_restart()

    def register_code_block(self, generator_function, name="", no_refresh=False):
        new_block = CodeBlock(self.owner, generator_function, name, no_refresh=no_refresh)
        self.code_blocks[new_block.name] = new_block
        print(f"New code block registered: {new_block.name}")
        return new_block

    def _update(self, dt):
        for name in self.code_blocks:
            self.current_block = self.code_blocks[name]
            self.code_blocks[name].update(dt)

    def clone(self, owner):
        clone = CodeManager(owner)
        for name in self.code_blocks:
            code_block = self.code_blocks[name]
            cloned_block = clone.register_code_block(code_block.generator_function, "clone", code_block.no_refresh)
            if code_block in self.cloned_blocks:
                clone.cloned_blocks.append(cloned_block)
        return clone

class CodeBlock():
    '''
    The CodeBlock encapsulates a generator and manages its state.

    
    '''
    last_id = -1

    def __init__(self, sprite_or_stage, generator_function, name="", no_refresh=False):
        '''
        Parameters
        ----------
        sprite_or_stage :
            The sprite or stage instance this code block is associated with
        generator_function : function
            The function containing the code to be executed. It can be a generator 
            function or a normal function which will be turned into a generator function
            if no_refresh is set to False (default)
        name : str, optional
            A name postfix used to distinguish this code block from other code blocks
            using the same function.
        no_refresh : bool, optional
            If set to true, the function must not be a generator function and will also
            not turned into one. This means that the screen will not be refreshed until
            the function returns.

        Raises
        ------
        ValueError
            If the function does not get a single parameter (for the reference to 
            the sprite or stage) or if a generator function is supplied with no_refresh 
            set to True. 
        '''
        if len(inspect.signature(generator_function).parameters) == 0:
            self.inject_instance = False
        elif len(inspect.signature(generator_function).parameters) == 1:
            self.inject_instance = True
        else:
            raise ValueError(
                f"Your code block '{generator_function.__name__}' needs zero or one parameter. The parameter is usually called self.")

        self.sprite_or_stage = sprite_or_stage
        if name != "" and name != None:
            name = f"-{name}"
        CodeBlock.last_id += 1
        # The name of a code block is unique with a global id.
        # A custom name may be provided to distinguish blocks
        # that use the same generator function.
        # This is not possible in Scratch (you only can copy blocks), 
        # but could be allowed here.
        # TODO: Discuss, if we should completely prohibit the 
        # reuse of generators, it could make things easier.
        self.name = f"{generator_function.__name__}{name}-{CodeBlock.last_id}"
        # A copy of the generator_function used for restarts (e.g. when bound to events)
        self.generator_function = generator_function
        # The current state of a generator after it is started
        self.generator = None
        self.is_function = False
        self.no_refresh = no_refresh
        if inspect.isgeneratorfunction(generator_function):
            if no_refresh:
                raise ValueError(
                    f"Your code block '{generator_function.__name__}' is set to no refresh. In this case, yield must not be used.")
        else:
            if no_refresh:
                # We support also plain functions, i.e. without yield
                self.is_function = True
            else:
                self.generator_function = self.add_yields(generator_function)

        # The time until the next step is executed
        self.wait_time = 0
        self.add_to_wait_time = 0
        self.gliding = False
        self.gliding_seconds = 0
        self.gliding_start_position = (0, 0)
        self.gliding_end_position = (0, 0)

        self.saying = False
        # Flag indicating if the block is currently running
        self.running = False
        # Ask mode (waiting for user input)
        self.asking = False

    def ask(self, question):
        # Go into "ask" mode
        self.asking = True
        # Queue the question in the central input manager
        self.sprite_or_stage.stage.input_manager.queue(question, self)


    def start_if_not_running(self):
        '''
        Start the code block. If the block is already started,
        this method does nothing.
        '''
        if self.running:
            return
        self.start_or_restart()

    def start_or_restart(self):
        '''
        (Re-)start the code block. If the block is already started,
        the current execution will not continue.
        '''
        self.running = True
        self.wait_time = 0
        if not self.is_function:
            target = self.sprite_or_stage
            if self.sprite_or_stage.facade:
                target = self.sprite_or_stage.facade
            if self.inject_instance:
                self.generator = self.generator_function(target)
            else:
                self.generator = self.generator_function()
        print(f"Start of {self.name} triggered.")

    def stop(self):
        """
        If there is other behaviors need to be stopped, add them here.
        """
        self.running = False
        self.asking = False
        self.saying = False
        if type(self.sprite_or_stage).__name__ == "CoreSprite":
            self.sprite_or_stage.bubble_manager.kill()
        if type(self.sprite_or_stage).__name__ == "CoreStage":
            self.sprite_or_stage.input_manager.cancel_asking()
        print(f"Stop of {self.name} triggered.")

    def update(self, dt):
        '''
        Check if it is time for the next step and execute it.
        '''
        if not self.running:
            return
        # Do nothing while waiting for an answer.
        if self.asking:
            return
        self.wait_time -= dt
        if self.wait_time < 0:
            if self.gliding:
                self.x, self.y = self.gliding_end_position
                self.gliding = False
            if self.saying:
                self.sprite_or_stage.bubble_manager.kill()
                self.saying = False
            if self.is_function:
                target = self.sprite_or_stage
                if self.sprite_or_stage.facade:
                    target = self.sprite_or_stage.facade
                self.generator = self.generator_function(target)
                print(f"CodeBlock {self.name} has finished.")
                self.running = False
                return
            else:
                try:
                    result = next(self.generator)
                    self.wait_time = 0
                    if isinstance(result, float) or isinstance(result, int):
                        self.wait_time = result
                    if self.add_to_wait_time > 0:
                        self.wait_time += self.add_to_wait_time
                        self.add_to_wait_time = 0
                except StopIteration:
                    print(f"CodeBlock {self.name} has finished.")
                    self.running = False
        elif self.gliding:
            self.sprite_or_stage.motion_setx(
                self.gliding_end_position[0] - (self.gliding_end_position[0] - self.gliding_start_position[0]) * (
                            self.wait_time / self.gliding_seconds))
            self.sprite_or_stage.motion_sety(
                self.gliding_end_position[1] - (self.gliding_end_position[1] - self.gliding_start_position[1]) * (
                            self.wait_time / self.gliding_seconds))

    def index_ast(self, func_ast):
        '''
        Helper function to (re-)generate for each node that is in
        a body or orelse field its parent, the field and its position 
        in the field.

        Parameters
        ----------
        func_ast : ast.AST
            The AST to be indexed.
        '''

        for node in ast.walk(func_ast):
            for field in ["body", "orelse"]:
                if hasattr(node, field):
                    for index, child in enumerate(getattr(node, field)):
                        child.parent = node
                        child.index = index
                        child.isin = field

    def add_yields(self, function):
        '''
        This function does the black magic and adds yields to the code so that
        the screen can update. This creates a generator function.

        Parameters
        ----------
        function : The function to be transformed to a generator function.
        '''
        func_ast = ast.parse(inspect.getsource(function))

        # yield at the end of the function
        func_ast.body[0].body.append(ast.Expr(value=ast.Yield(value=ast.Constant(value=0))))

        self.index_ast(func_ast)

        for node in ast.walk(func_ast):

            # yield at the end of for and while iterations
            if isinstance(node, ast.For):
                node.body.append(ast.Expr(value=ast.Yield(value=ast.Constant(value=0))))
            if isinstance(node, ast.While):
                node.body.append(ast.Expr(value=ast.Yield(value=ast.Constant(value=0))))

            # yield after certain calls defined in CodeBlock.yield_funcs
            if isinstance(node, ast.Expr) and isinstance(node.value, ast.Call) and isinstance(node.value.func,
                                                                                              ast.Attribute):
                corefunc = get_core_function_from_instance(node.value.func.attr, self.sprite_or_stage.facade)
                if corefunc in yield_funcs:
                    getattr(node.parent, node.isin).insert(node.index + 1,
                                                           ast.Expr(value=ast.Yield(value=ast.Constant(value=0))))
                    # We need to reindex so that further yields get the right position
                    self.index_ast(func_ast)

        ast.fix_missing_locations(func_ast)
        # print(ast.unparse(func_ast)) # outputs the transformed code
        namespace = {}
        namespace.update(function.__globals__)
        code = compile(func_ast, "<string>", mode="exec")
        exec(code, namespace)
        return namespace[function.__name__]
    
    def __str__(self):
        return f"<CodeBlock {self.name} - {self.running}>"
    
    def __repr__(self):
        return str(self)



from pystage.en import Sprite, Stage

stage = Stage()
zombie1 = stage.add_a_sprite()
zombie2 = stage.add_a_sprite()

def z1(zombie: Sprite):
    zombie.go_to_x_y(-40, 0)
    zombie.broadcast("hello")


def z2(zombie: Sprite):
    zombie.go_to_x_y(40, 0)
    zombie.say("You said hello!")


zombie1.when_program_starts(z1)
zombie2.when_i_receive_message("hello", z2)

stage.play()



from pystage.core.sprite import CoreSprite
from pystage.core.stage import CoreStage

stage = CoreStage()
stage.pystage_addbackdrop("grid")
sprite = stage.pystage_createsprite()



def printpos(self):
    print("Pos", self.motion_xposition(), self.motion_yposition())
    print("PG Pos", self._pos)
    print("Center", self.costume_manager.get_center())
    print("Rect", self.rect)

def turn(self):
    self.motion_turnleft(45)

def move_rect(self):
    self.motion_changexby(10)

sprite.event_whenkeypressed(" ", printpos)
sprite.event_whenkeypressed("t", turn)
sprite.event_whenkeypressed("m", move_rect)

stage.pystage_play()



from pystage.en import Stage

stage = Stage()
zombie = stage.add_a_sprite()

def doit():
    zombie.ask_and_wait("Hi, how are you?")
    zombie.say(f"Great, that you are {zombie.answer()}!")
    zombie.wait(2)
    zombie.ask_and_wait("And what else?")
    zombie.ask_and_wait("Just one more question... what is 7+7?")
    zombie.say(f"Your answer: {zombie.answer()}")

zombie.when_program_starts(doit)

stage.play()



from pystage.core.sprite import CoreSprite
from pystage.core.stage import CoreStage

stage = CoreStage()
stage.pystage_addbackdrop("grid")
sprite = stage.pystage_createsprite()
sprite.pystage_addsound("dancehead2")



def do_something(self: CoreSprite):
    self.looks_say("Hello pyStage!")
    for i in range(4):
        self.looks_changesizeby(20)
        self.motion_movesteps(20)
        self.control_wait(1)
        # self.looks_think("This is awesome!")
        self.motion_turnleft(90)
        self.control_wait(1)
        # self.looks_think("")
    self.looks_say("Move me around with WASD.")

sprite.event_whenflagclicked(do_something)


def soundcheck(self: CoreSprite):
    self.sound_playuntildone("dancehead2")
    self.looks_say("Sound finished!")

sprite.event_whenflagclicked(soundcheck)

def right(self: CoreSprite):
    self.motion_changexby(10)

def left(self: CoreSprite):
    self.motion_changexby(-10)

def up(self: CoreSprite):
    self.motion_changeyby(10)

def down(self: CoreSprite):
    self.motion_changeyby(-10)

def mouse(self: CoreSprite):
    self.motion_pointtowards_pointer()
    self.looks_say(f"Mouse pos: {self.sensing_mousex()} / {self.sensing_mousey()}")

def say_space_pressed(self: CoreSprite):
    while True:
        if self.sensing_keypressed(" "):
            self.looks_say("Space pressed!")

def pen_toggle(self: CoreSprite):
    if self.pen:
        self.pen_penUp()
    else:
        self.pen_penDown()

def pen_clear(self: CoreSprite):
    self.pen_clear()

sprite.event_whenkeypressed("d", right)
sprite.event_whenkeypressed("a", left)
sprite.event_whenkeypressed("w", up)
sprite.event_whenkeypressed("s", down)
sprite.event_whenkeypressed("m", mouse)
sprite.event_whenkeypressed("p", pen_toggle)
sprite.event_whenkeypressed("c", pen_clear)
sprite.event_whenflagclicked(say_space_pressed)

stage.pystage_play()



from pystage.en import Stage

stage = Stage()
stage.add_backdrop("grid")
zombie = stage.add_a_sprite()

def doit(zombie):
    zombie.say("Hello pyStage!")
    for i in range(4):
        zombie.move(10)
        zombie.wait(1)
        zombie.think("This is awesome!")
        zombie.turn_left(90)
        zombie.wait(1)
        zombie.think("")
    zombie.say("Move me around with WASD.")

zombie.when_program_starts(doit)

def right(zombie):
    zombie.change_x_by(10)

def left(zombie):
    zombie.change_x_by(-10)

def up(zombie):
    zombie.change_y_by(10)

def down(zombie):
    zombie.change_y_by(-10)

zombie.when_key_pressed("d", right)
zombie.when_key_pressed("a", left)
zombie.when_key_pressed("w", up)
zombie.when_key_pressed("s", down)

stage.play()



from pystage.en import Stage

stage = Stage()
zombie = stage.add_a_sprite()
zombie.set_ghost_effect_to(20)

def doit():
    for i in range(4):
        zombie.move(10)
        zombie.turn_left(90)
        zombie.wait(1)
        zombie.change_ghost_effect_by(20)
        print(zombie.y_position())

zombie.when_program_starts(doit)

stage.play()



import importlib
import inspect
import dis

def get_translated_function(corefunc, language, stage=False):
    """Get translated API function for a core function name.

    Use the function metadata to determine the correct translated API
    function based on a given core function name.

    Parameters
    ----------
    corefunc : str
        A core function name, i.e. a method from CoreSprite or CoreStage
    language : str
        The language of the target API
    stage : bool
        True, if the stage API is to be used, otherwise defaults to sprite.
    """
    if language=="core":
        return corefunc
    lang = importlib.import_module(f"pystage.{language}")
    cls = lang.stage_class if stage else lang.sprite_class
    for name, func in inspect.getmembers(cls, predicate=inspect.isfunction):
        for i in dis.Bytecode(func):
            if (i.opname=="LOAD_METHOD" or i.opname=="LOAD_ATTR") and i.argval==corefunc:
                # print(f"Translated: {name}")
                return name
    # print(f"No translated API method found for {corefunc}")
    return None


def get_core_function(translated, language, stage=False):
    """Get core API function for a translated function name.

    Use the function metadata to determine the correct translated API
    function based on a given core function name.

    Parameters
    ----------
    translated : str
        A translated function name, i.e. a method from a sprite or stage class
    language : str
        The language of the source API
    stage : bool
        True, if the stage API is to be used, otherwise defaults to sprite.
    """
    if language=="core":
        return translated
    lang = importlib.import_module(f"pystage.{language}")
    cls = lang.stage_class if stage else lang.sprite_class
    for name, func in inspect.getmembers(cls, predicate=inspect.isfunction):
        for i in dis.Bytecode(func):
            if (i.opname=="LOAD_METHOD" or i.opname=="LOAD_ATTR") and name==translated and i.argval!="_core":
                return i.argval
    return None

def get_core_function_from_instance(translated, instance):
    """Get core API function for a translated function name.

    Use the function metadata to determine the correct translated API
    function based on a given core function name.

    Parameters
    ----------
    translated : str
        A translated function name, i.e. a method from a sprite or stage class
    instance : object
        A stage or sprite instance from a translated API
    """
    # If core is used directly, return translated:
    if not instance:
        return translated
    cls = instance.__class__
    for name, func in inspect.getmembers(cls, predicate=inspect.isfunction):
        for i in dis.Bytecode(func):
            if (i.opname=="LOAD_METHOD" or i.opname=="LOAD_ATTR") and name==translated and i.argval!="_core":
                return i.argval
    return None




from pystage.core.gui import BubbleManager, Bubble
from pystage.core.assets import CostumeManager
from pystage.core._base_sprite import BaseSprite


class _Looks(BaseSprite):
    def __init__(self):
        super().__init__()

    def looks_switchbackdropto(self, backdrop):
        # Backdrops are for the stage.
        # In Scratch, a sprite can change the backdrop.
        self.stage.costume_manager.switch_costume(backdrop)

    def looks_nextbackdrop(self):
        self.stage.costume_manager.next_costume()

    def looks_seteffectto_color(self, value):
        pass

    looks_seteffectto_color.opcode="looks_seteffectto"
    looks_seteffectto_color.param="EFFECT"
    looks_seteffectto_color.value="COLOR"
    looks_seteffectto_color.translation="looks_effect_color"


    def looks_seteffectto_fisheye(self, value):
        pass
    
    looks_seteffectto_fisheye.opcode="looks_seteffectto"
    looks_seteffectto_fisheye.param="EFFECT"
    looks_seteffectto_fisheye.value="FISHEYE"
    looks_seteffectto_fisheye.translation="looks_effect_fisheye"


    def looks_seteffectto_whirl(self, value):
        pass
    
    looks_seteffectto_whirl.opcode="looks_seteffectto"
    looks_seteffectto_whirl.param="EFFECT"
    looks_seteffectto_whirl.value="WHIRL"
    looks_seteffectto_whirl.translation="looks_effect_whirl"


    def looks_seteffectto_pixelate(self, value):
        pass
    
    looks_seteffectto_pixelate.opcode="looks_seteffectto"
    looks_seteffectto_pixelate.param="EFFECT"
    looks_seteffectto_pixelate.value="PIXELATE"
    looks_seteffectto_pixelate.translation="looks_effect_pixelate"


    def looks_seteffectto_mosaic(self, value):
        pass
    
    looks_seteffectto_mosaic.opcode="looks_seteffectto"
    looks_seteffectto_mosaic.param="EFFECT"
    looks_seteffectto_mosaic.value="MOSAIC"
    looks_seteffectto_mosaic.translation="looks_effect_mosaic"
    

    def looks_seteffectto_brightness(self, value):
        pass
    
    looks_seteffectto_brightness.opcode="looks_seteffectto"
    looks_seteffectto_brightness.param="EFFECT"
    looks_seteffectto_brightness.value="BRIGHTNESS"
    looks_seteffectto_brightness.translation="looks_effect_brightness"
    

    def looks_seteffectto_ghost(self, value):
        self.ghost = value
    
    looks_seteffectto_ghost.opcode="looks_seteffectto"
    looks_seteffectto_ghost.param="EFFECT"
    looks_seteffectto_ghost.value="GHOST"
    looks_seteffectto_ghost.translation="looks_effect_ghost"
    

    def looks_changeeffectby_color(self, value):
        pass

    looks_changeeffectby_color.opcode="looks_changeeffectby"
    looks_changeeffectby_color.param="EFFECT"
    looks_changeeffectby_color.value="COLOR"
    looks_changeeffectby_color.translation="looks_effect_color"


    def looks_changeeffectby_fisheye(self, value):
        pass
    
    looks_changeeffectby_fisheye.opcode="looks_changeeffectby"
    looks_changeeffectby_fisheye.param="EFFECT"
    looks_changeeffectby_fisheye.value="FISHEYE"
    looks_changeeffectby_fisheye.translation="looks_effect_fisheye"


    def looks_changeeffectby_whirl(self, value):
        pass
    
    looks_changeeffectby_whirl.opcode="looks_changeeffectby"
    looks_changeeffectby_whirl.param="EFFECT"
    looks_changeeffectby_whirl.value="WHIRL"
    looks_changeeffectby_whirl.translation="looks_effect_whirl"


    def looks_changeeffectby_pixelate(self, value):
        pass
    
    looks_changeeffectby_pixelate.opcode="looks_changeeffectby"
    looks_changeeffectby_pixelate.param="EFFECT"
    looks_changeeffectby_pixelate.value="PIXELATE"
    looks_changeeffectby_pixelate.translation="looks_effect_pixelate"


    def looks_changeeffectby_mosaic(self, value):
        pass
    
    looks_changeeffectby_mosaic.opcode="looks_changeeffectby"
    looks_changeeffectby_mosaic.param="EFFECT"
    looks_changeeffectby_mosaic.value="MOSAIC"
    looks_changeeffectby_mosaic.translation="looks_effect_mosaic"
    

    def looks_changeeffectby_brightness(self, value):
        pass
    
    looks_changeeffectby_brightness.opcode="looks_changeeffectby"
    looks_changeeffectby_brightness.param="EFFECT"
    looks_changeeffectby_brightness.value="BRIGHTNESS"
    looks_changeeffectby_brightness.translation="looks_effect_brightness"
    

    def looks_changeeffectby_ghost(self, value):
        self.ghost += value
    
    looks_changeeffectby_ghost.opcode="looks_changeeffectby"
    looks_changeeffectby_ghost.param="EFFECT"
    looks_changeeffectby_ghost.value="GHOST"
    looks_changeeffectby_ghost.translation="looks_effect_ghost"
    

    def looks_cleargraphiceffects(self):
        pass

    def looks_backdropnumbername_number(self):
        # 1-based
        return self.stage.costume_manager.current_costume + 1

    looks_backdropnumbername_number.opcode="looks_backdropnumbername"
    looks_backdropnumbername_number.param="NUMBER_NAME"
    looks_backdropnumbername_number.value="number"
    looks_backdropnumbername_number.translation="looks_numbername_number"

    looks_backdropnumbername_number.return_type = int


    def looks_backdropnumbername_name(self):
        costume = self.stage.costume_manager.get_costume()
        if not costume:
            return ""
        return costume.name

    looks_backdropnumbername_name.opcode="looks_backdropnumbername"
    looks_backdropnumbername_name.param="NUMBER_NAME"
    looks_backdropnumbername_name.value="name"
    looks_backdropnumbername_name.translation="looks_numbername_name"

    looks_backdropnumbername_name.return_type = str

class _LooksSprite(_Looks):

    def __init__(self):
        super().__init__()
        self.bubble_manager = BubbleManager(self)
        # Make sure to clone all properties in sprite.py that need to be cloned
        self.size = 100
        self.ghost = 0 # 0 - visible / 100 - transparent
        self.visible = True


    def pystage_addcostume(self, name, center_x=None, center_y=None, factor=1):
        self.costume_manager.add_costume(name, center_x, center_y, factor)

    def pystage_replacecostume(self, index, name, center_x=None, center_y=None, factor=1):
        self.costume_manager.replace_costume(index, name, center_x, center_y, factor)

    def pystage_insertcostume(self, index, name, center_x=None, center_y=None, factor=1):
        self.costume_manager.insert_costume(index, name, center_x, center_y, factor)

    def looks_sayforsecs(self, text, secs):
        self.bubble_manager.sayforsecs(text, secs)


    def looks_say(self, text):
        self.bubble_manager.say(text)


    def looks_thinkforsecs(self, text, secs):
        self.bubble_manager.thinkforsecs(text, secs)

    def looks_think(self, text):
        self.bubble_manager.think(text)

    def looks_switchcostumeto(self, costume):
        self.costume_manager.switch_costume(costume)

    def looks_nextcostume(self):
        self.costume_manager.next_costume()


    def looks_changesizeby(self, percent):
        # this is percentage
        self.size += percent


    def looks_setsizeto(self, percent):
        self.size = percent

    def looks_show(self):
        print("Show", self)
        if self.visible:
            return
        self.visible = True
        self.stage._update_visible()


    def looks_hide(self):
        print ("Hide", self)
        if not self.visible:
            return
        self.visible = False
        self.stage._update_visible()


    def looks_gotofrontback_front(self):
        self.stage.sprites.to_front(self)

    looks_gotofrontback_front.opcode="looks_gotofrontback"
    looks_gotofrontback_front.param="FRONT_BACK"
    looks_gotofrontback_front.value="front"

    
    def looks_gotofrontback_back(self):
        self.stage.sprites.to_back(self)

    looks_gotofrontback_back.opcode="looks_gotofrontback"
    looks_gotofrontback_back.param="FRONT_BACK"
    looks_gotofrontback_back.value="back"


    def looks_goforwardbackwardlayers_forward(self, value):
        self.stage.sprites.layer_forward(self, value)

    looks_goforwardbackwardlayers_forward.opcode="looks_goforwardbackwardlayers"
    looks_goforwardbackwardlayers_forward.param="FORWARD_BACKWARD"
    looks_goforwardbackwardlayers_forward.value="forward"

    def looks_goforwardbackwardlayers_backward(self, value):
        self.stage.sprites.layer_backward(self, value)

    looks_goforwardbackwardlayers_backward.opcode="looks_goforwardbackwardlayers"
    looks_goforwardbackwardlayers_backward.param="FORWARD_BACKWARD"
    looks_goforwardbackwardlayers_backward.value="backward"


    def looks_costumenumbername_number(self):
        # 1-based
        return self.costume_manager.current_costume + 1

    looks_costumenumbername_number.opcode="looks_costumenumbername"
    looks_costumenumbername_number.param="NUMBER_NAME"
    looks_costumenumbername_number.value="number"
    looks_costumenumbername_number.translation="looks_numbername_number"

    looks_costumenumbername_number.return_type = int


    def looks_costumenumbername_name(self):
        costume = self.costume_manager.get_costume()
        if not costume:
            return ""
        return costume.name

    looks_costumenumbername_name.opcode="looks_costumenumbername"
    looks_costumenumbername_name.param="NUMBER_NAME"
    looks_costumenumbername_name.value="name"
    looks_costumenumbername_name.translation="looks_numbername_name"

    looks_costumenumbername_name.return_type = str

    def looks_size(self):
        # percent
        return self.size

    looks_size.return_type = float


class _LooksStage(_Looks):

    def __init__(self):
        super().__init__()


    def pystage_addbackdrop(self, name, center_x=None, center_y=None):
        self.costume_manager.add_costume(name, center_x, center_y)


    def pystage_replacebackdrop(self, index, name, center_x=None, center_y=None):
        self.costume_manager.replace_costume(index, name, center_x, center_y)


    def pystage_insertbackdrop(self, index, name, center_x=None, center_y=None):
        self.costume_manager.insert_costume(index, name, center_x, center_y)


    def looks_switchbackdroptoandwait(self, backdrop):
        # Wait for any code under a "when backdrop is changed"
        pass






import random
import math

class _Operators():
    def __init__(self):
        super().__init__()

    def operator_random(self, start, end):
        return random.randint(start, end)

    def operator_mathop(self, operator, number): 
        ops = {
                'abs': abs,
                'floor': math.floor,
                'ceiling': math.ceil,
                'sqrt': math.sqrt,
                'sin': lambda n: math.sin((math.pi * n) / 180),
                'cos': lambda n: math.cos((math.pi * n) / 180),
                'tan': lambda n: math.tan((math.pi * n) / 180) if not n == 90 else float("inf"),
                'asin': lambda n: (math.asin(n) * 180) / math.pi,
                'acos': lambda n: (math.acos(n) * 180) / math.pi,
                'atan': lambda n: (math.atan(n) * 180) / math.pi,
                'ln': lambda n : math.log(n) if not n == 0 else -float("inf"),
                'log': lambda n: math.log10(n) if not n == 0 else -float("inf"),
                'e ^': math.exp,
                '10 ^': lambda n: math.pow(10, n),
                }
        if operator not in ops:
            raise ValueError(f"Operator {operator} not supported.")
        
        try:
            value = ops[operator](number)
            if value == float("inf"):
                return "Infinity"
            elif value == -float("inf"):
                return "-Infinity"
            # avoid for situations like sin(30) == 0.49999999999999994
            return round(value, 10)
        except ValueError:
            # print(f"ValueError: math domain error: {operator}({number})")
            return "NaN"
        except OverflowError:
            return "Infinity"



from pystage.en import Stage

stage = Stage()
pen = stage.add_a_sprite("pen")

def doit():
    pen.pen_down()
    for i in range(4):
        pen.move(10)
        pen.turn_left(90)
        pen.wait(1)
        print(pen.y_position())

pen.when_program_starts(doit)

stage.play()



import argparse
import inspect
import json
import requests
import re
import textwrap
import importlib

from pystage.core.sprite import CoreSprite
from pystage.core.stage import CoreStage
from pystage.l10n.translations import trans, funcname


square_bracketed_variable = re.compile(r"\[[^\]]*\]")
percent_variable = re.compile(r"%[0-9]")
to_underscore = re.compile(r"[ -:/,%]")
to_delete = re.compile(r"[%?]")
multiple_underscores = re.compile(r"_+")
deleteself = re.compile(r"self[, ]*")
deletedefaults = re.compile(r"=[^),]+")

# Functions from PyGame sprite class
ignore_function_names = ["update", "add", "add_internal", "alive", "groups", "kill", "remove", "remove_internal"]

def get_translations(lang="en"):
    blocks = json.loads(requests.get(f"https://raw.githubusercontent.com/LLK/scratch-l10n/master/editor/blocks/{lang}.json").text)
    extensions = json.loads(requests.get(f"https://raw.githubusercontent.com/pystage/scratch-l10n/master/editor/extensions/{lang}.json").text)
    res = {}
    for key in blocks:
        res[key.upper()] = blocks[key] 
    for key in extensions:
        res[key.upper().replace(".", "_")] = extensions[key] 
    res["MOTION_TURNLEFT"] = res["MOTION_TURNLEFT"].replace("%1", res["BOOST_TILTDIRECTION_LEFT"])
    res["MOTION_TURNRIGHT"] = res["MOTION_TURNRIGHT"].replace("%1", res["BOOST_TILTDIRECTION_RIGHT"])
    res["EVENT_WHENFLAGCLICKED"] = res["EVENT_WHENFLAGCLICKED"].replace("%1", "<greenflag>")
    res["EVENT_WHENGREATERTHAN"] = res["EVENT_WHENGREATERTHAN"].replace(">", "<greater>")
    # Another inconsistency where not the opcode is used (typo?)...
    res["SOUND_SETEFFECTTO"] = res["SOUND_SETEFFECTO"]
    return res


def get_translation(func, translations, lang="en"):
    name = func.__name__
    opcode = name.upper()
    outer_key = None
    trans_key = name.upper()
    trans_key = trans_key.replace("OPERATOR", "OPERATORS")
    if hasattr(func, "opcode"):
        opcode = func.opcode.upper()
        if name.upper() in translations:
            outer_key = opcode
            trans_key = name.upper()
        else:
            trans_key = opcode
    if hasattr(func, "translation"):
        trans_key = func.translation.upper()
    trans = ""
    if trans_key in translations:
        trans = translations[trans_key]
    if opcode != trans_key and opcode in translations:
        outer_key = opcode
    if hasattr(func, "outer_translation"):
        outer_key = func.outer_translation.upper()
    if outer_key:
        trans = translations[outer_key]
        position = "%1"
        if hasattr(func, "position"):
            position = func.position
        if position in trans:
            trans = trans.replace(position, translations[trans_key])
        else:
            trans += f" {translations[trans_key]}"
    return trans


def create_funcname(translation, translations):
    translation = square_bracketed_variable.sub("", translation)
    translation = to_underscore.sub("_", translation)
    translation = percent_variable.sub("", translation)
    translation = to_delete.sub("", translation)
    translation = translation.replace("#", translations["LOOKS_NUMBERNAME_NUMBER"])
    translation = multiple_underscores.sub("_", translation)
    translation = translation.strip("_").lower()
    translation = translation.replace("<greenflag>", "GREENFLAG")
    translation = translation.replace("<greater>", "GREATERTHAN")
    return translation

if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        prog="python -m pystage.l10n",
        description="Manage translations")
    parser.add_argument("-t", "--translations", action="store_true", help="List core function translations")
    parser.add_argument("-a", "--api", action="store_true", help="List translated API function suggestions")
    parser.add_argument("-s", "--stage", action="store_true", help="Use stage instead of sprite")
    parser.add_argument("-l", "--language", metavar="LANG", type=str, help="API language, 2-letter ISO code, e.g. en, de, ...", nargs="?", default="en")
    args = parser.parse_args()

    if args.translations:
        cls = CoreStage if args.stage else CoreSprite
        translations = get_translations(args.language)
        for name, func in inspect.getmembers(cls, predicate=inspect.isfunction):
            if name.startswith("_"):
                continue
            params = [key for key in inspect.signature(func).parameters]
            params.remove("self")
            params = f"({', '.join(params)})"
            trans = get_translation(func, translations, lang=args.language)
            print(f"{name}{params} - {trans}\n")
    elif args.api:
        cls = CoreStage if args.stage else CoreSprite
        language = args.language
        sprite_class = trans("gui.backpack.spriteLabel", language).capitalize()
        stage_class = trans("gui.stageSelector.stage", language).capitalize()

        create_sprite = funcname(trans("gui.howtos.animate-char.step_addsprite", language), language)
        add_sound = funcname(trans("gui.howtos.animate-char.step_addsound", language), language)
        play = funcname(trans("gui.playButton.play", language), language)
            
        if args.stage:
            print(f'''
from pystage.core.stage import CoreStage
from pystage.{language}.sprite import {sprite_class}


class {stage_class}():

    def __init__(self):
        self._core = CoreStage()
        self._core.facade = self
        self._core.sprite_facade_class = {sprite_class}

    def {create_sprite}(self, costume="default"):
        return self._core.pystage_createsprite(costume=costume)

    def {play}(self):
        self._core.pystage_play()
        
            ''')
        else:
            print(f'''
from pystage.core.sprite import CoreSprite


class {sprite_class}():
    def __init__(self, core_sprite):
        self._core : CoreSprite = core_sprite
        self._core.facade = self


            ''')
        translations = get_translations(args.language)
        for name, func in inspect.getmembers(cls, predicate=inspect.isfunction):
            if name.startswith("_") or name in ignore_function_names:
                continue
            param_keys = [key for key in inspect.signature(func).parameters]
            params = [str(inspect.signature(func).parameters[key]) for key in param_keys]
            params_call = f"({', '.join(params)})"
            trans_text = get_translation(func, translations, lang=args.language)
            trans_hint = ""
            if hasattr(func, "translation_hint"):
                trans_hint = f"""
        TODO TRANSLATORS: {func.translation_hint}
        """
            trans = create_funcname(trans_text, translations)
            if not trans:
                trans = name
            paramdoc = "\n        ".join([f"{param.split('=')[0]} : FILL" for param in params if param != "self"])
            if paramdoc:
                paramdoc = f'''
        Parameters
        ----------
        {paramdoc}
        '''
            print(f'''\
    def {trans}{params_call}:
        """{trans_text}{trans_hint}

        Translation string: {trans_text}
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        {paramdoc}

        Returns
        -------

        """
        return self._core.{name}{deletedefaults.sub("", deleteself.sub("", params_call))}
                ''')






'''
Rules:
    - Every template ends at the indent_level where it started.
    - At the end of block code add a blank line.
    - New functions (hat blocks) add a newline at the beginning to get to 2 blank lines.
'''

templates = {
        
        "event_whenflagclicked": '''\
                
                def {{func}}_{{ID}}(self):
                    {{NEXT | indent(4) }}

                {{CURRENT_SPRITE}}.{{func}}({{func}}_{{ID}})

                ''',
        
        "event_whenbroadcastreceived": '''\

                def {{func}}_{{ID}}(self):
                    {{NEXT | indent(4)}}

                {{CURRENT_SPRITE}}.{{func}}({{BROADCAST_OPTION}}, {{func}}_{{ID}})

                ''',

        "event_whenkeypressed": '''\

                def {{func}}_{{ID}}(self):
                    {{NEXT | indent(4)}}

                {{CURRENT_SPRITE}}.{{func}}({{KEY_OPTION}}, {{func}}_{{ID}})
                
                ''',

        "event_whenthisspriteclicked": '''\

                def {{func}}_{{ID}}(self):
                    {{NEXT | indent(4)}}

                {{CURRENT_SPRITE}}.{{func}}({{func}}_{{ID}})
                
                ''',

        "event_whengreaterthan": '''\
                def {{func}}_{{ID}}(self):
                    {{NEXT | indent(4)}}
                {{CURRENT_SPRITE}}.{{func}}({{VALUE}}, {{func}}_{{ID}})
                
                ''',

        "event_whenbackdropswitchesto": '''\
                def {{func}}_{{ID}}(self):
                    {{NEXT | indent(4)}}
                {{CURRENT_SPRITE}}.{{func}}({{BACKDROP | global_backdrop}}, {{func}}_{{ID}})
                
                ''',

        "control_start_as_clone": '''\

                def {{func}}_{{ID}}(self):
                    {{NEXT | indent(4)}}

                {{CURRENT_SPRITE}}.{{func}}({{func}}_{{ID}})
                
                ''',

        "control_forever":'''\
                while True:
                    {{SUBSTACK | indent(4)}}
                ''',

        "control_repeat":'''\
                for _ in range({{TIMES}}):
                    {{SUBSTACK | indent(4)}}
                ''',

        "control_repeat_until": '''\
                while not {{CONDITION}}:
                    {{SUBSTACK | indent(4)}}
                ''',
        
        "control_wait_until": '''\
                while not {{CONDITION}}:
                    pass
                ''',

        "control_if": '''\
                if {{CONDITION}}:
                    {{SUBSTACK | indent(4)}}
                ''',

        "control_if_else":'''\
                if {{CONDITION}}:
                    {{SUBSTACK | indent(4)}}
                else:
                    {{SUBSTACK2 | indent(4)}}
                ''',

        "operator_add": "({{NUM1}} + {{NUM2}})",
        "operator_and": "({{OPERAND1}} and {{OPERAND2}})",
        "operator_contains": "({{STRING2}} in {{STRING1}})",
        "operator_divide": "({{NUM1}} / {{NUM2}})",
        "operator_equals": "({{OPERAND1}} == {{OPERAND2}})",
        "operator_gt": "({{OPERAND1}} > {{OPERAND2}})",
        "operator_join": '"".join([{{STRING1}}, {{STRING2}}])',
        "operator_length": "len({{STRING}})",
        "operator_letter_of": "{{STRING}}[{{LETTER}}-1]",
        "operator_lt": "({{OPERAND1}} < {{OPERAND2}})",
        "operator_mod": "({{NUM1}} % {{NUM2}})",
        "operator_multiply": "({{NUM1}} * {{NUM2}})",
        "operator_not": "not ({{OPERAND}})",
        "operator_or": "({{OPERAND1}} or {{OPERAND2}})",
        "operator_round": "round({{NUM}})",
        "operator_subtract": "({{NUM1}} - {{NUM2}})",

        "looks_costume": "{{COSTUME | global_costume}}",
        "looks_backdrops": "{{BACKDROP | global_backdrop}}",
        "sound_sounds_menu": "{{SOUND_MENU | global_sound }}",

        "sensing_keyoptions": "{{KEY_OPTION}}",
        "sensing_distancetomenu": "{{DISTANCETOMENU | global_sprite}}",
        "sensing_touchingobjectmenu": "{{TOUCHINGOBJECTMENU | global_sprite}}",
        "control_create_clone_of_menu": "{{CLONE_OPTION | global_sprite }}",
        "sensing_of_object_menu": "{{OBJECT | global_sprite }}",
        "motion_pointtowards_menu": "{{TOWARDS | global_sprite }}",
                }





import requests
import json
import re

_translations = {}

square_bracketed_variable = re.compile(r"\[[^\]]*\]")
percent_variable = re.compile(r"%[0-9]")
to_underscore = re.compile(r"[ -:/,%]")
to_delete = re.compile(r"[%?]")
multiple_underscores = re.compile(r"_+")
deleteself = re.compile(r"self[, ]*")


def fetch_translations(lang="en"):
    if lang in _translations:
        return _translations[lang]

    blocks = json.loads(requests.get(f"https://raw.githubusercontent.com/LLK/scratch-l10n/master/editor/blocks/{lang}.json").text)
    extensions = json.loads(requests.get(f"https://raw.githubusercontent.com/pystage/scratch-l10n/master/editor/extensions/{lang}.json").text)
    interface = json.loads(requests.get(f"https://raw.githubusercontent.com/pystage/scratch-l10n/master/editor/interface/{lang}.json").text)
    res = {}
    for key in blocks:
        res[key.upper()] = blocks[key] 
    for key in extensions:
        res[key.upper().replace(".", "_")] = extensions[key] 
    for key in interface:
        res[key.upper()] = interface[key] 
    res["MOTION_TURNLEFT"] = res["MOTION_TURNLEFT"].replace("%1", res["BOOST_TILTDIRECTION_LEFT"])
    res["MOTION_TURNRIGHT"] = res["MOTION_TURNRIGHT"].replace("%1", res["BOOST_TILTDIRECTION_RIGHT"])
    res["EVENT_WHENFLAGCLICKED"] = res["EVENT_WHENFLAGCLICKED"].replace("%1", "<greenflag>")
    res["EVENT_WHENGREATERTHAN"] = res["EVENT_WHENGREATERTHAN"].replace(">", "<greater>")
    # Another inconsistency where not the opcode is used (typo?)...
    res["SOUND_SETEFFECTTO"] = res["SOUND_SETEFFECTO"]
    _translations[lang] = res
    return res

def trans(key, lang):
    translations = fetch_translations(lang)
    key = key.upper()
    if key in translations:
        return translations[key]
    return f"MISSING_TRANSLATION: {key}"

def funcname(translation, lang):
    translation = square_bracketed_variable.sub("", translation)
    translation = to_underscore.sub("_", translation)
    translation = percent_variable.sub("", translation)
    translation = to_delete.sub("", translation)
    translation = translation.replace("#", trans("LOOKS_NUMBERNAME_NUMBER", lang))
    translation = multiple_underscores.sub("_", translation)
    translation = translation.strip("_").lower()
    translation = translation.replace("<greenflag>", "GREENFLAG")
    translation = translation.replace("<greater>", "GREATERTHAN")
    return translation



import textwrap
import re
import logging
import inspect
import importlib
import dis
from jinja2 import Environment
from jinja2.exceptions import UndefinedError
logger = logging.getLogger(__name__)

from pystage.core.sprite import CoreSprite
from pystage.core.stage import CoreStage

to_underscore = re.compile(r"[^A-Za-z0-9_]")
valid_variable_start = re.compile(r"[a-zA-Z_]")
multiple_underscores = re.compile(r"_+")


def quoted(value):
    if not isinstance(value, str):
        return value
    if value.startswith("\""):
        return value
    else:
        return f"\"{value}\""


def unquoted(value):
    if not isinstance(value, str):
        return value
    if value.startswith("\""):
        return value.strip("\"")
    else:
        return value

def resolve(value):
    # print(f"Resolving: {value}")
    if isinstance(value, str):
        return value
    for key in value["params"]:
        res = resolve(value["params"][key])
        # print(f"Resolved value: {res}")
        return res
    raise ValueError(f"No suitable value found: {value}")




class CodeWriter():
    '''
    Helper class to write code templates
    '''
    def __init__(self, project, templates, language="core"):
        super().__init__()
        self.project = project
        self.templates = templates
        self.language = language

        self.comments = []
        self.last_id = 0
        self.current_sprite = ""
        self.jinja_environment = Environment(trim_blocks=True, lstrip_blocks=True)
        self.jinja_environment.filters["global_sound"] = lambda name: self.global_sound(name)
        self.jinja_environment.filters["global_costume"] = lambda name: self.global_costume(name)
        self.jinja_environment.filters["global_backdrop"] = lambda name: self.global_backdrop(name)
        self.jinja_environment.filters["global_sprite"] = lambda name: self.get_sprite_var(unquoted(name))
        
        logger.debug("CodeWriter created.")


    def set_sprite(self, name):
        self.current_sprite = name


    def get_sprite_var(self, name=None):
        def to_python(name: str):
            vname = name.lower().strip()
            # Replace everything that is not allowed in a 
            # variable name with underscores
            vname = to_underscore.sub("_", vname)
            vname = multiple_underscores.sub("_", vname)
            # Check of the variable does not start with a number
            if not valid_variable_start.fullmatch(vname[0]):
                vname = "_" + vname
            return vname
        if name is None:
            return to_python(self.current_sprite)
        else:
            return to_python(name)

    def get_sprite_or_stage(self, name=None):
        if not name:
            name = self.current_sprite
        if self.project["stage"]["name"] == name:
            return self.project["stage"]
        for sprite in self.project["sprites"]:
            if sprite["name"]==name:
                return sprite
        raise ValueError(f"No stage or sprite found with name '{name}'.")

    def get_id(self):
        """Helper to create unique names.

        """
        self.last_id += 1
        return self.last_id


    def get_opcode_function(self, block):
        """Get core API function for a block.

        Use the function metadata to determine the correct core API
        function based on a given block, using opcode and params.

        Parameters
        ----------
        block : dict
            A block from the intermediate code representation.
        """
        # print("Getting function for opcode: " + block["opcode"])
        cls = CoreStage if block["stage"] else CoreSprite
        # print(f"Searching in class: {cls}")
        elsefunc = None
        for name, func in inspect.getmembers(cls, predicate=inspect.isfunction):
            # print(f"Testing: {name} - {func}")
            if name==block["opcode"]:
                # print(f"Matching API method: {name}")
                return func
            if hasattr(func, "opcode"):
                if func.opcode == block["opcode"]:
                    if hasattr(func, "param"):
                        if func.param in block["params"]:
                            value = unquoted(resolve(block["params"][func.param]))
                            if func.value==value:
                                # print(f"Matching API method: {name}")
                                return func
                    else:
                        elsefunc = func
        if elsefunc:
            # print(f"Matching API method: {elsefunc}")
            return elsefunc
        print(f"No API method for {block.opcode}")
        return None


    def get_translated_function(self, block, language):
        """Get translated API function for a block.

        Use the function metadata to determine the correct translated API
        function based on a given block, using opcode and params.

        Parameters
        ----------
        block : dict
            A block from the intermediate code representation.
        """
        corefunc = self.get_opcode_function(block)
        if language=="core":
            return corefunc
        if corefunc is None:
            return None
        lang = importlib.import_module(f"pystage.{language}")
        cls = lang.stage_class if block["stage"] else lang.sprite_class
        for name, func in inspect.getmembers(cls, predicate=inspect.isfunction):
            for i in dis.Bytecode(func):
                if (i.opname=="LOAD_METHOD" or i.opname=="LOAD_ATTR") and i.argval==corefunc.__name__:
                    # print(f"Translated: {name}")
                    return func
        # print(f"No translated API method found for {block.opcode}")
        return None
            
        
    def get_translated_call(self, block, language):
        corefunc = self.get_opcode_function(block)
        func = self.get_translated_function(block, language)
        if func is None:
            return quoted(f"NO TRANSLATION: {block.opcode}")
        res = func.__name__ + "("
        fieldname = corefunc.param if hasattr(corefunc, "param") else None
        res += ", ".join([str(block.params[p]) for p in block.params if p != fieldname])
        res += ")"
        # print(res)
        return res


    def get_translated_template(self, block, language):
        corefunc = self.get_opcode_function(block)
        func = self.get_translated_function(block, language)
        if func is None:
            return quoted(f"NO TRANSLATION: {block.opcode}")
        res = f"self.{func.__name__}("
        fieldname = corefunc.param if hasattr(corefunc, "param") else None
        res += ", ".join(["{{" + p + "}}" for p in block.params if p != fieldname])
        res += ")"
        # print(res)
        return res



    def render_comments(self):
        res = ""
        for c in self.comments:
            for line in c.split("\n"):
                res += f"# {line}\n"
        self.comments.clear()
        return res


    def global_sound(self, name: str, quoted=True):
        name = unquoted(name)
        sprite = self.get_sprite_or_stage()
        for sound in sprite["sounds"]:
            if sound["local_name"] == name:
                q = '"' if quoted else ''
                return f'{q}{self.project["sounds"][sound["md5"]]["global_name"]}{q}'
        raise ValueError(f"No sound with name '{name}' found for sprite '{sprite['name']}'")
        

    def global_costume(self, name, quoted=True):
        name = unquoted(name)
        sprite = self.get_sprite_or_stage()
        for costume in sprite["costumes"]:
            if costume["local_name"] == name:
                q = '"' if quoted else ''
                return f'{q}{self.project["costumes"][costume["md5"]]["global_name"]}{q}'
        raise ValueError(f"No costume with name '{name}' found for sprite '{sprite['name']}'")


    def global_backdrop(self, name, quoted=True):
        name = unquoted(name)
        sprite = self.project["stage"]
        for costume in sprite["costumes"]:
            if costume["local_name"] == name:
                q = '"' if quoted else ''
                return f'{q}{self.project["costumes"][costume["md5"]]["global_name"]}{q}'
        raise ValueError(f"No backdrop with name '{name}' found for stage.")


    def process(self, block):
        if not isinstance(block, dict):
            # We have a simple value
            return str(block)
        else:
            # First we add comments to the queue
            if "comments" in block:
                self.comments.extend(block["comments"])
            # If the block has its own template, we simply render it.
            if block["opcode"] in self.templates:
                template = self.templates[block["opcode"]]
                context = {}
                # if the template needs a translated function name, we deliver it
                if "{{func}}" in template:
                    func = self.get_translated_function(block, self.language)
                    context["func"] = func.__name__ if func is not None else f"<<NO_FUNCTION-{block['opcode']}>>"
                return self.render(block, template, context)
            else:
                # We use the default template mechanism
                default_template = self.get_translated_template(block, self.language)
                return self.render(block, default_template)


    def render(self, block, text, context={}):
        text = textwrap.dedent(text)
        if block["next"]:
            context["NEXT"] = self.process(block["next"])
            if not "NEXT" in text:
                text += "\n{{NEXT}}"
        for param in block["params"]:
            context[param] = self.process(block["params"][param])
        context["CURRENT_SPRITE"] = self.get_sprite_var()
        if "{{ID}}" in text:
            context["ID"] = self.get_id()

        # if {{NEXT | indent (4)}} is still in text, but there is no
        # NEXT in context, then it means the Scratch “hat” block is empty,
        # so set NEXT to "pass", and the function is essentially empty.
        if "indent(4)" in text:
            if "NEXT" in text and not context.get("NEXT"):
                context["NEXT"] = "pass"
            if "SUBSTACK" in text and not context.get("SUBSTACK"):
                context["SUBSTACK"] = "pass"
            if "SUBSTACK2" in text and not context.get("SUBSTACK2"):
                context["SUBSTACK2"] = "pass"
        if "{{CONDITION}}" in text and not context.get("CONDITION"):
            context["CONDITION"] = "None"
        template = self.jinja_environment.from_string(text)
        try:
            text = template.render(context)
        except UndefinedError as e:
            print(f"Template variable not available: {e}")
        return text




from pystage.core.constants import KEY_MAPPINGS
from pystage.core._base_sprite import BaseSprite


class _Events(BaseSprite):

    def __init__(self):
        super().__init__()
    ##
    # Events
    #

    def event_whenflagclicked(self, generator_function, name="", no_refresh=False):
        new_block = self.code_manager.register_code_block(generator_function, name)
        print(f"Bound to start: {new_block.name}")
        new_block.start_or_restart()


    def event_whenkeypressed(self, key, generator_function, name="", no_refresh=False):
        '''
        Adds the code block to the event queue for key presses.
        '''
        new_block = self.code_manager.register_code_block(generator_function, name, no_refresh)
        if key not in KEY_MAPPINGS:
            # TODO: implement "any" key.
            raise ValueError(f"Bad key: {key}. Only a-z, 0-9 and space are allowed.")
        pg_key = KEY_MAPPINGS[key]
        # No defaultdict so that we can easily check if a key mapping is available
        if pg_key not in self.code_manager.key_pressed_blocks:
            self.code_manager.key_pressed_blocks[pg_key] = []
        self.code_manager.key_pressed_blocks[pg_key].append(new_block.name)
        print(f"Bound to key press ({key}/{pg_key}): {new_block.name}")


    def event_whenthisspriteclicked(self, generator_function, name="", no_refresh=False):
        '''
        Adds the code block to the event queue for clicks.
        '''
        new_block = self.code_manager.register_code_block(generator_function, name, no_refresh)
        self.code_manager.clicked_blocks.append(new_block)
        print(f"Bound to click: {new_block.name}")



    def event_whenbackdropswitchesto(self, backdrop, generator_function, name="", no_refresh=False):
        pass


    def event_whengreaterthan_loudness(self, value, generator_function, name="", no_refresh=False):
        # Not sure if this can/should be implemented, requires microphone access.
        # See sensing for more information.
        print("Microphone access is currently not supported. See https://github.com/pystage/pystage/issues/116")
    event_whengreaterthan_loudness.opcode="event_whengreaterthan"
    event_whengreaterthan_loudness.param="WHENGREATERTHANMENU"
    event_whengreaterthan_loudness.value="LOUDNESS"


    def event_whengreaterthan_timer(self, value, generator_function, name="", no_refresh=False):
        # Scratch has a timer that can be reset. 
        pass
    event_whengreaterthan_timer.opcode="event_whengreaterthan"
    event_whengreaterthan_timer.param="WHENGREATERTHANMENU"
    event_whengreaterthan_timer.value="TIMER"


    def event_whenbroadcastreceived(self, message, generator_function, name="", no_refresh=False):
        '''
        Adds the code block to the event queue for broadcasts.
        '''
        new_block = self.code_manager.register_code_block(generator_function, name, no_refresh)
        # No defaultdict so that we can easily check if a mapping is available
        if message not in self.code_manager.broadcast_blocks:
            self.code_manager.broadcast_blocks[message] = []
        self.code_manager.broadcast_blocks[message].append(new_block.name)
        print(f"Bound to broadcast message '{message}': {new_block.name}")


    def event_broadcast(self, message):
        self.stage.message_broker.broadcast(message)


    def event_broadcastandwait(self, message):
        # waits until all receiver scripts finish. Tricky.
        pass




from pystage.core.sprite import CoreSprite


class Sprite:
    def __init__(self, core_sprite):
        self._core: CoreSprite = core_sprite
        self._core.facade = self

    def create_clone_of_sprite(self, sprite="_myself_"):
        """create clone of %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.control_create_clone_of_sprite(sprite)

    def create_clone_of_myself(self):
        """create clone of myself

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.control_create_clone_of_myself()
        
    def delete_this_clone(self):
        """delete this clone

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.control_delete_this_clone()

    def when_i_start_as_a_clone(
        self, generator_function, name="", no_refresh=False
    ):
        """when I start as a clone

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.control_start_as_clone(
            generator_function, name, no_refresh
        )

    def stop_all(self):
        """stop all

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.control_stop_all()

    def stop_other_scripts_in_sprite(self):
        """stop other scripts in sprite

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.control_stop_other()

    def stop_this_script(self):
        """stop this script

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.control_stop_this()

    def wait(self, secs):
        """wait %1 seconds

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        secs : FILL
        

        Returns
        -------

        """
        return self._core.control_wait(secs)

    def change_variable_by(self, name, value):
        """change %1 by %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        value : FILL
        

        Returns
        -------

        """
        return self._core.data_changevariableby(name, value)

    def hide_variable(self, name):
        """hide variable %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.data_hidevariable(name)

    def set_variable(self, name, value):
        """set %1 to %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        value : FILL
        

        Returns
        -------

        """
        return self._core.data_setvariableto(name, value)

    def show_variable(self, name):
        """show variable %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.data_showvariable(name)

    def get_variable(self, name):
        """

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.data_variable(name)

    def broadcast(self, message):
        """broadcast %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        message : FILL
        

        Returns
        -------

        """
        return self._core.event_broadcast(message)

    def broadcast_and_wait(self, message):
        """broadcast %1 and wait

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        message : FILL
        

        Returns
        -------

        """
        return self._core.event_broadcastandwait(message)

    def when_backdrop_switches_to(
        self, backdrop, generator_function, name="", no_refresh=False
    ):
        """when backdrop switches to %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        backdrop : FILL
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.event_whenbackdropswitchesto(
            backdrop, generator_function, name, no_refresh
        )

    def when_i_receive_message(
        self, message, generator_function, name="", no_refresh=False
    ):
        """when I receive %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        message : FILL
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.event_whenbroadcastreceived(
            message, generator_function, name, no_refresh
        )

    def when_program_starts(self, generator_function, name="", no_refresh=False):
        """when <greenflag> clicked

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.event_whenflagclicked(
            generator_function, name, no_refresh
        )

    def when_loudness_greater_than(
        self, value, generator_function, name="", no_refresh=False
    ):
        """when loudness <greater> %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.event_whengreaterthan_loudness(
            value, generator_function, name, no_refresh
        )

    def when_timer_greater_than(
        self, value, generator_function, name="", no_refresh=False
    ):
        """when timer <greater> %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.event_whengreaterthan_timer(
            value, generator_function, name, no_refresh
        )

    def when_key_pressed(self, key, generator_function, name="", no_refresh=False):
        """when %1 key pressed

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        key : FILL
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.event_whenkeypressed(
            key, generator_function, name, no_refresh
        )

    def when_this_sprite_clicked(self, generator_function, name="", no_refresh=False):
        """when this sprite clicked

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.event_whenthisspriteclicked(
            generator_function, name, no_refresh
        )

    def backdrop_name(self):
        """backdrop name

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_backdropnumbername_name()

    def backdrop_number(self):
        """backdrop number

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_backdropnumbername_number()

    def change_brightness_effect_by(self, value):
        """change brightness effect by %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_changeeffectby_brightness(value)

    def change_color_effect_by(self, value):
        """change color effect by %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_changeeffectby_color(value)

    def change_fisheye_effect_by(self, value):
        """change fisheye effect by %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_changeeffectby_fisheye(value)

    def change_ghost_effect_by(self, value):
        """change ghost effect by %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_changeeffectby_ghost(value)

    def change_mosaic_effect_by(self, value):
        """change mosaic effect by %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_changeeffectby_mosaic(value)

    def change_pixelate_effect_by(self, value):
        """change pixelate effect by %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_changeeffectby_pixelate(value)

    def change_whirl_effect_by(self, value):
        """change whirl effect by %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_changeeffectby_whirl(value)

    def change_size_by(self, percent):
        """change size by %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        percent : FILL
        

        Returns
        -------

        """
        return self._core.looks_changesizeby(percent)

    def clear_graphic_effects(self):
        """clear graphic effects

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_cleargraphiceffects()

    def costume_name(self):
        """costume name

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_costumenumbername_name()

    def costume_number(self):
        """costume number

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_costumenumbername_number()

    def go_backward(self, value):
        """go backward %2 layers

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_goforwardbackwardlayers_backward(value)

    def go_forward(self, value):
        """go forward %2 layers

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_goforwardbackwardlayers_forward(value)

    def go_to_back_layer(self):
        """go to back layer

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_gotofrontback_back()

    def go_to_front_layer(self):
        """go to front layer

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_gotofrontback_front()

    def hide(self):
        """hide

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_hide()

    def next_backdrop(self):
        """next backdrop

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_nextbackdrop()

    def next_costume(self):
        """next costume

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_nextcostume()

    def say(self, text):
        """say %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        text : FILL
        

        Returns
        -------

        """
        return self._core.looks_say(text)

    def say_for_seconds(self, text, secs):
        """say %1 for %2 seconds

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        text : FILL
        secs : FILL
        

        Returns
        -------

        """
        return self._core.looks_sayforsecs(text, secs)

    def set_brightness_effect_to(self, value):
        """set brightness effect to %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_seteffectto_brightness(value)

    def set_color_effect_to(self, value):
        """set color effect to %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_seteffectto_color(value)

    def set_fisheye_effect_to(self, value):
        """set fisheye effect to %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_seteffectto_fisheye(value)

    def set_ghost_effect_to(self, value):
        """set ghost effect to %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_seteffectto_ghost(value)

    def set_mosaic_effect_to(self, value):
        """set mosaic effect to %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_seteffectto_mosaic(value)

    def set_pixelate_effect_to(self, value):
        """set pixelate effect to %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_seteffectto_pixelate(value)

    def set_whirl_effect_to(self, value):
        """set whirl effect to %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_seteffectto_whirl(value)

    def set_size_to(self, percent):
        """set size to %1 %

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        percent : FILL
        

        Returns
        -------

        """
        return self._core.looks_setsizeto(percent)

    def show(self):
        """show

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_show()

    def size(self):
        """size

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_size()

    def switch_backdrop_to(self, backdrop):
        """switch backdrop to %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        backdrop : FILL
        

        Returns
        -------

        """
        return self._core.looks_switchbackdropto(backdrop)

    def switch_costume_to(self, costume):
        """switch costume to %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        costume : FILL
        

        Returns
        -------

        """
        return self._core.looks_switchcostumeto(costume)

    def think(self, text):
        """think %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        text : FILL
        

        Returns
        -------

        """
        return self._core.looks_think(text)

    def think_for_seconds(self, text, secs):
        """think %1 for %2 seconds

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        text : FILL
        secs : FILL
        

        Returns
        -------

        """
        return self._core.looks_thinkforsecs(text, secs)

    def change_x_by(self, value):
        """change x by %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.motion_changexby(value)

    def change_y_by(self, value):
        """change y by %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.motion_changeyby(value)

    def direction(self):
        """direction

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.motion_direction()

    def glide_to_x_y(self, secs, x, y):
        """glide %1 secs to x: %2 y: %3

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        secs : FILL
        x : FILL
        y : FILL
        

        Returns
        -------

        """
        return self._core.motion_glidesecstoxy(secs, x, y)

    def glide_to_mouse_pointer(self, secs):
        """glide %1 secs to mouse-pointer

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        secs : FILL
        

        Returns
        -------

        """
        return self._core.motion_glideto_pointer(secs)

    def glide_to_random_position(self, secs):
        """glide %1 secs to random position

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        secs : FILL
        

        Returns
        -------

        """
        return self._core.motion_glideto_random(secs)

    def glide_to_sprite(self, secs, sprite):
        """glide %1 secs to %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        secs : FILL
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.motion_glideto_sprite(secs, sprite)

    def go_to_mouse_pointer(self):
        """go to mouse-pointer

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.motion_goto_pointer()

    def go_to_random_position(self):
        """go to random position

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.motion_goto_random()

    def go_to_sprite(self, sprite):
        """go to %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.motion_goto_sprite(sprite)

    def go_to_x_y(self, x, y):
        """go to x: %1 y: %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        x : FILL
        y : FILL
        

        Returns
        -------

        """
        return self._core.motion_gotoxy(x, y)

    def if_on_edge_bounce(self):
        """if on edge, bounce

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.motion_ifonedgebounce()

    def move(self, steps):
        """move %1 steps

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        steps : FILL
        

        Returns
        -------

        """
        return self._core.motion_movesteps(steps)

    def point_in_direction(self, direction):
        """point in direction %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        direction : FILL
        

        Returns
        -------

        """
        return self._core.motion_pointindirection(direction)

    def point_towards_mouse_pointer(self):
        """point towards mouse-pointer

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.motion_pointtowards_pointer()

    def point_towards_sprite(self, sprite):
        """point towards %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.motion_pointtowards_sprite(sprite)

    def set_rotation_style_none(self):
        """set rotation style %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        """
        return self._core.motion_setrotationstyle_dontrotate()

    def set_rotation_style_left_right(self):
        """set rotation style %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...


        """
        return self._core.motion_setrotationstyle_leftright()

    def set_rotation_style_around(self):
        """set rotation style %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        """
        return self._core.motion_setrotationstyle_allaround()

    def set_x(self, value):
        """set x to %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.motion_setx(value)

    def set_y(self, value):
        """set y to %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.motion_sety(value)

    def turn_left(self, deg):
        """turn left %2 degrees

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        deg : FILL
        

        Returns
        -------

        """
        return self._core.motion_turnleft(deg)

    def turn_right(self, deg):
        """turn right %2 degrees

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        deg : FILL
        

        Returns
        -------

        """
        return self._core.motion_turnright(deg)

    def x_position(self):
        """x position

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.motion_xposition()

    def y_position(self):
        """y position

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.motion_yposition()

    def calculate(self, operator, number):
        """%1 of %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        operator : FILL
        number : FILL
        

        Returns
        -------

        """
        return self._core.operator_mathop(operator, number)

    def pick_random(self, start, end):
        """pick random %1 to %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        start : FILL
        end : FILL
        

        Returns
        -------

        """
        return self._core.operator_random(start, end)

    def change_pen_brightness_by(self, value):
        """change pen brightness by [VALUE]

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.pen_changePenColorParamBy_brightness(value)

    def change_pen_hue_by(self, value):
        """change pen color by [VALUE]

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.pen_changePenColorParamBy_color(value)

    def change_pen_saturation_by(self, value):
        """change pen saturation by [VALUE]

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.pen_changePenColorParamBy_saturation(value)

    def change_pen_transparency_by(self, value):
        """change pen transparency by [VALUE]

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.pen_changePenColorParamBy_transparency(value)

    def change_pen_size_by(self, value):
        """change pen size by [SIZE]

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.pen_changePenSizeBy(value)

    def erase_all(self):
        """erase all

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.pen_clear()

    def pen_down(self):
        """pen down

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.pen_penDown()

    def pen_up(self):
        """pen up

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.pen_penUp()

    def set_pen_brightness_to(self, value):
        """set pen brightness to [VALUE]

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.pen_setPenColorParamTo_brightness(value)

    def set_pen_hue_to(self, value):
        """set pen color to [VALUE]

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.pen_setPenColorParamTo_color(value)

    def set_pen_saturation_to(self, value):
        """set pen saturation to [VALUE]

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.pen_setPenColorParamTo_saturation(value)

    def set_pen_transparency_to(self, value):
        """set pen transparency to [VALUE]

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.pen_setPenColorParamTo_transparency(value)

    def set_pen_color(self, color):
        """set pen color to [COLOR]

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        color : FILL
        

        Returns
        -------

        """
        return self._core.pen_setPenColorToColor(color)

    def set_pen_size_to(self, value):
        """set pen size to [SIZE]

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.pen_setPenSizeTo(value)

    def stamp(self):
        """stamp

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.pen_stamp()

    def add_costume(self, name, center_x=None, center_y=None, factor=1):
        """

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        center_x : FILL
        center_y : FILL
        factor : FILL
        

        Returns
        -------

        """
        return self._core.pystage_addcostume(name, center_x, center_y, factor)

    def add_sound(self, name):
        """

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.pystage_addsound(name)

    def insert_costume(self, index, name, center_x=None, center_y=None, factor=1):
        """

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        index : FILL
        name : FILL
        center_x : FILL
        center_y : FILL
        

        Returns
        -------

        """
        return self._core.pystage_insertcostume(index, name, center_x, center_y, factor)

    def create_variable(self, name, all_sprites=True):
        """

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        all_sprites : FILL
        

        Returns
        -------

        """
        return self._core.pystage_makevariable(name, all_sprites=True)

    def replace_costume(self, index, name, center_x=None, center_y=None, factor=1):
        """

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        index : FILL
        name : FILL
        center_x : FILL
        center_y : FILL
        

        Returns
        -------

        """
        return self._core.pystage_replacecostume(
            index, name, center_x, center_y, factor
        )

    def set_monitor_position(self, name, x, y):
        """

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        x : FILL
        y : FILL
        

        Returns
        -------

        """
        return self._core.pystage_setmonitorposition(name, x, y)

    def pystage_setmonitorstyle_large(self, name):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.pystage_setmonitorstyle_large(name)

    def pystage_setmonitorstyle_normal(self, name):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.pystage_setmonitorstyle_normal(name)

    def pystage_setmonitorstyle_slider(self, name):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.pystage_setmonitorstyle_slider(name)

    def answer(self):
        """answer

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_answer()

    def ask_and_wait(self, question):
        """ask %1 and wait

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        question : FILL
        

        Returns
        -------

        """
        return self._core.sensing_askandwait(question)

    def color_is_touching(self, sprite_color, color):
        """color %1 is touching %2?

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite_color : FILL
        color : FILL
        

        Returns
        -------

        """
        return self._core.sensing_coloristouchingcolor(sprite_color, color)

    def current_date(self):
        """current date

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_current_date()

    def current_day_of_week(self):
        """current day of week

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_current_dayofweek()

    def current_hour(self):
        """current hour

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_current_hour()

    def current_minute(self):
        """current minute

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_current_minute()

    def current_month(self):
        """current month

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_current_month()

    def current_second(self):
        """current second

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_current_second()

    def current_year(self):
        """current year

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_current_year()

    def days_since(self):
        """days since 2000

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_dayssince2000()

    def distance_to_mouse_pointer(self):
        """distance to mouse-pointer

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_distanceto_pointer()

    def distance_to_sprite(self, sprite):
        """distance to %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_distanceto_sprite(sprite)

    def key_pressed(self, key):
        """key %1 pressed?

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        key : FILL
        

        Returns
        -------

        """
        return self._core.sensing_keypressed(key)

    def loudness(self):
        """loudness

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_loudness()

    def mouse_down(self):
        """mouse down?

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_mousedown()

    def mouse_x(self):
        """mouse x

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_mousex()

    def mouse_y(self):
        """mouse y

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_mousey()

    def backdrop_name_of(self, stage="_stage_"):
        """backdrop name of %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        stage : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_backdropname(stage="_stage_")

    def backdrop_of(self, stage="_stage_"):
        """backdrop # of %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        stage : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_backdropnumber(stage="_stage_")

    def costume_name_of(self, sprite):
        """costume name of %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_costumename(sprite)

    def costume_of(self, sprite):
        """costume # of %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_costumenumber(sprite)

    def direction_of(self, sprite):
        """direction of %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_direction(sprite)

    def size_of(self, sprite):
        """size of %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_size(sprite)

    def get_variable_of(self, variable, sprite="_stage_"):
        """%1 of %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        variable : FILL
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_variable(variable, sprite="_stage_")

    def volume_of(self, sprite="_stage_"):
        """volume of %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_volume(sprite="_stage_")

    def x_position_of(self, sprite):
        """x position of %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_xposition(sprite)

    def y_position_of(self, sprite):
        """y position of %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_yposition(sprite)

    def reset_timer(self):
        """reset timer

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_resettimer()

    def set_drag_mode_draggable(self):
        """set drag mode draggable

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_setdragmode_draggable()

    def set_drag_mode_not_draggable(self):
        """set drag mode not draggable

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_setdragmode_notdraggable()

    def timer(self):
        """timer

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_timer()

    def touching_color(self, color):
        """touching color %1?

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        color : FILL
        

        Returns
        -------

        """
        return self._core.sensing_touchingcolor(color)

    def touching_edge(self):
        """touching edge?

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_touchingobject_edge()

    def touching_mouse_pointer(self):
        """touching mouse-pointer?

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_touchingobject_pointer()

    def touching(self, sprite):
        """touching %1?

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_touchingobject_sprite(sprite)

    def username(self):
        """username

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_username()

    def change_pan_left_right_effect_by(self, value):
        """change pan left/right effect by %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.sound_changeeffectby_pan(value)

    def change_pitch_effect_by(self, value):
        """change pitch effect by %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.sound_changeeffectby_pitch(value)

    def change_volume_by(self, value):
        """change volume by %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.sound_changevolumeby(value)

    def clear_sound_effects(self):
        """clear sound effects

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sound_cleareffects()

    def start_sound(self, name, loop=0):
        """start sound %1

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        loop : FILL
        

        Returns
        -------

        """
        return self._core.sound_play(name, loop=0)

    def play_sound_until_done(self, name):
        """play sound %1 until done

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.sound_playuntildone(name)

    def set_pan_left_right_effect_to(self, value):
        """set pan left/right effect to %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.sound_seteffectto_pan(value)

    def set_pitch_effect_to(self, value):
        """set pitch effect to %2

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.sound_seteffectto_pitch(value)

    def set_volume_to(self, value):
        """set volume to %1%

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.sound_setvolumeto(value)

    def stop_all_sounds(self):
        """stop all sounds

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sound_stopallsounds()

    def volume(self):
        """volume

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sound_volume()

    def set_name(self, name):
        """set name of the sprite to %1%

        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.pystage_setname(name)



import pygame
import pkg_resources

light_font_9 = pygame.font.Font(pkg_resources.resource_filename("pystage", "fonts/roboto-light.ttf"), 10)

class InputManager:

    def __init__(self, stage):
        self._queue = []
        self._answer = ""
        self._stage = stage

    @property
    def answer(self):
        return self._answer
    
    def queue(self, text, codeblock):
        question = Question(self, text, codeblock)
        self._queue.append(question)

    def is_active(self):
        return len(self._queue) > 0

    def accept_answer(self, answer):
        self._answer = answer
        del self._queue[0]

    def cancel_asking(self):
        if self._queue:
            self._queue[0].codeblock.asking = False
            del self._queue[0]
    
    def process_key(self, event):
        if self.is_active():
            self._queue[0].process_key(event)

    def update(self, dt):
        if len(self._queue) == 0:
            return
        question = self._queue[0]
        question.update(dt)


    def draw(self, surface):
        if len(self._queue) == 0:
            return
        question = self._queue[0]
        question.draw(surface)


class Question:

    OUTER_MARGIN = 5
    OUTER_HEIGHT = 55
    INNER_MARGIN = 13
    INNER_HEIGHT = 25
    GLOW = 3
    GLOW_COLOR = pygame.Color(193, 219, 255)
    ACTIVE_COLOR = pygame.Color(76, 151, 255)
    BLINK_TIME = 1

    def __init__(self, manager, text, codeblock):
        self.manager = manager
        self.text = text
        self.codeblock = codeblock
        self.answer = ""
        self.active = False
        self.stage = self.codeblock.sprite_or_stage.stage
        self.is_stage_question = self.codeblock.sprite_or_stage == self.stage
        self.blink_time = Question.BLINK_TIME
        self.cursor_visible = True
        self.rendered_answer = pygame.Surface((0,0),flags=pygame.SRCALPHA)

    def process_key(self, event):
        key = event.key
        # No CTRL stuff, leads to unknown characters
        if pygame.key.get_mods() & pygame.KMOD_CTRL:
            return
        # backspace and delete both work like backspace
        elif key == pygame.K_BACKSPACE or key == pygame.K_DELETE:
            if len(self.answer) > 0:
                self.answer = self.answer[:-1]
        # Return ends the input
        elif key == pygame.K_RETURN:
            self.manager.accept_answer(self.answer)
            self.codeblock.asking = False
            if not self.is_stage_question:
                self.codeblock.sprite_or_stage.looks_say("")
        # Add unicode representation of the pressed key to the answer
        else:
            self.answer += event.unicode
        # Update rendered text
        rendered = light_font_9.render(self.answer, True, pygame.Color(0,0,0))
        text_width = self.stage.width - 2 * Question.OUTER_MARGIN - 2 * Question.INNER_MARGIN - 36 
        cutoff = max(rendered.get_width() - text_width, 0)
        self.rendered_answer = rendered.subsurface((cutoff, 0), (min(rendered.get_width(), text_width), rendered.get_height()))

    def update(self, dt):
        if not self.active:
            self.active = True
            pygame.key.set_repeat(500, 50)
            pygame.key.start_text_input()
            if not self.is_stage_question:
                self.codeblock.sprite_or_stage.looks_say(self.text)
        self.blink_time -= dt
        if self.blink_time < 0:
            self.blink_time = Question.BLINK_TIME
            self.cursor_visible = not self.cursor_visible

    def draw(self, surface: pygame.Surface):
        area = pygame.Rect(
                Question.OUTER_MARGIN, 
                self.stage.height - Question.OUTER_HEIGHT - Question.OUTER_MARGIN, 
                self.stage.width - 2 * Question.OUTER_MARGIN, 
                Question.OUTER_HEIGHT)
        field = pygame.Rect(
                Question.OUTER_MARGIN + 2 + Question.INNER_MARGIN, 
                self.stage.height - Question.INNER_HEIGHT - Question.OUTER_MARGIN - 2 - Question.INNER_MARGIN, 
                self.stage.width - 2 * Question.OUTER_MARGIN - 2* Question.INNER_MARGIN - 4, 
                Question.INNER_HEIGHT)
        glow = pygame.Rect(field.x - Question.GLOW, field.y - Question.GLOW, field.width + 2*Question.GLOW, field.height + 2*Question.GLOW)
        pygame.draw.rect(surface, pygame.Color(200,200,200), area, 2, 5)
        pygame.draw.rect(surface, Question.GLOW_COLOR, glow, 0, 20)
        pygame.draw.rect(surface, pygame.Color(255,255,255), field, 0, 20)
        pygame.draw.rect(surface, Question.ACTIVE_COLOR, field, 1, 20)
        surface.blit(self.rendered_answer, (field.x + 15, field.y + 6))
        if self.cursor_visible:
            pygame.draw.line(surface, pygame.Color(0,0,0), (field.x + 15 + self.rendered_answer.get_width(), field.y + 8),(field.x + 15 + self.rendered_answer.get_width(), field.y + field.height - 8), 1)




from pystage.core._base_sprite import BaseSprite
from pystage.core.monitors import Monitor

class _Variables(BaseSprite):
    # Also monitors

    # Implementation needs to be aware of naming conflicts and raise Exceptions if a name
    # is used both globally and locally.

    # global variables are stored in stage (stage variables are global in Scratch)
    # local variables are stored in the sprite
    # dictionaries are used to store the variables

    def __init__(self):
        super().__init__()

        self.variables = {}
        self.monitors = {}


    def _get_monitor(self, name):
        if name in self.monitors:
            return self.monitors[name]
        elif name in self.stage.monitors:
            return self.stage.monitors[name]
        return None


    def data_setvariableto(self, name, value):
        if name in self.variables:
            self.variables[name] = value
        elif name in self.stage.variables:
            self.stage.variables[name] = value
        else:
            raise ValueError(f"The variable {name} does not exist.")


    def data_variable(self, name):
        if name in self.variables:
            return self.variables[name]
        elif name in self.stage.variables:
            return self.stage.variables[name]
        else:
            raise ValueError(f"The variable {name} does not exist.")


    def data_changevariableby(self, name, value):
        if name in self.variables:
            self.variables[name] += value
        elif name in self.stage.variables:
            self.stage.variables[name] += value
        else:
            raise ValueError(f"The variable {name} does not exist.")

    def data_showvariable(self, name):
        # Use smart positioning or old position, if we have one.
        monitor = self._get_monitor(name)
        if not monitor:
            return
        monitor.show()


    def data_hidevariable(self, name):
        monitor = self._get_monitor(name)
        if not monitor:
            return
        monitor.hide()


    def pystage_makevariable(self, name, all_sprites=True):
        # Make sure a variable name is unique for a sprite or globally. 
        # Same name for local variables is allowed!
        if name in self.stage.variables:
            raise ValueError(f"The variable {name} already exists!")
        if name in self.variables:
            raise ValueError(f"The variable {name} already exists!")
        if all_sprites:
            for sprite in self.stage.sprites:
                if name in sprite.variables:
                    raise ValueError(f"The variable {name} already exists!")

        monitor = Monitor(self, name)
        monitor.hide()
        monitor.set_function(lambda: self.data_variable(name))

        if not all_sprites:
            self.variables[name]=0
            self.monitors[name] = monitor
        else:
            self.stage.variables[name]=0
            self.stage.monitors[name] = monitor


    def pystage_setmonitorposition(self, name, x, y):
        monitor = self._get_monitor(name)
        if not monitor:
            return
        monitor.set_position(x, y)


    def pystage_setmonitorstyle_large(self, name):
        monitor = self._get_monitor(name)
        if not monitor:
            return
        monitor.set_style(Monitor.LARGE)


    def pystage_setmonitorstyle_normal(self, name):
        monitor = self._get_monitor(name)
        if not monitor:
            return
        monitor.set_style(Monitor.NORMAL)


    def pystage_setmonitorstyle_slider(self, name):
        monitor = self._get_monitor(name)
        if not monitor:
            return
        monitor.set_style(Monitor.SLIDER)



import pygame


KEY_MAPPINGS = {
        "a": pygame.K_a,
        "b": pygame.K_b,
        "c": pygame.K_c,
        "d": pygame.K_d,
        "e": pygame.K_e,
        "f": pygame.K_f,
        "g": pygame.K_g,
        "h": pygame.K_h,
        "i": pygame.K_i,
        "j": pygame.K_j,
        "k": pygame.K_k,
        "l": pygame.K_l,
        "m": pygame.K_m,
        "n": pygame.K_n,
        "o": pygame.K_o,
        "p": pygame.K_p,
        "q": pygame.K_q,
        "r": pygame.K_r,
        "s": pygame.K_s,
        "t": pygame.K_t,
        "u": pygame.K_u,
        "v": pygame.K_v,
        "w": pygame.K_w,
        "x": pygame.K_x,
        "y": pygame.K_y,
        "z": pygame.K_z,
        "1": pygame.K_1,
        "2": pygame.K_2,
        "3": pygame.K_3,
        "4": pygame.K_4,
        "5": pygame.K_5,
        "6": pygame.K_6,
        "7": pygame.K_7,
        "8": pygame.K_8,
        "9": pygame.K_9,
        "0": pygame.K_0,
        "left": pygame.K_LEFT,
        "right": pygame.K_RIGHT,
        "up": pygame.K_UP,
        "down": pygame.K_DOWN,
        "space": pygame.K_SPACE,
        " ": pygame.K_SPACE,
        }




class MessageBroker():
    """Utility class for message handling.

    The message broker belongs to the stage as central element.
    """

    def __init__(self, stage):
        self.stage = stage
        self._messages = []


    def broadcast(self, message):
        self._messages.append(message)


    def get_messages(self):
        return self._messages


    def mark_completed(self):
        self._messages.clear()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import shutil
import sys
from tempfile import TemporaryFile
from urllib.request import urlopen
from zipfile import ZipFile
import inspect
from docutils.transforms import Transform
from docutils import nodes

sys.path.insert(0, os.path.abspath("../../src"))


# -- Project information -----------------------------------------------------

project = 'PyStage'
copyright = '2021, The Pystage Developing Team'
author = 'The Pystage Developing Team'

# The full version, including alpha/beta/rc tags
release = '0.1'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc', 'sphinx.ext.napoleon', "sphinx_rtd_theme", "m2r2", "sphinx_panels"]
# 'sphinx.ext.coverage', 'sphinx.ext.napoleon'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# alabaster
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# doing this as example for en
PATH_BLOCK_IMAGES = "/".join(["_static", "images", "blocks"])
LOADED_LANG = []


# download block img
def download_pngs(lang):
    if lang not in LOADED_LANG:
        # first cleanup
        extract_path = "/".join(["source", PATH_BLOCK_IMAGES, lang])
        if os.path.exists(extract_path):
            return
            # shutil.rmtree(extract_path)

        block_img_url = f"http://img.pystage.org/blocks/zip/png/300/{lang}_png300.zip"

        with urlopen(block_img_url) as url:
            print(f"Downloading block imgs for {lang}.")
            html = url.read()

            with TemporaryFile() as tmp:
                tmp.write(html)
                with ZipFile(tmp) as f:
                    f.extractall(extract_path)

                    # rename all files: add lang name in file name to avoid same file names. Sphinx will pack all pngs
                    # in one directory
                    for file in os.listdir(extract_path):
                        try:
                            os.rename("/".join([extract_path, file]), "/".join([extract_path, f"{lang}_{file}"]))
                        except WindowsError as e:
                            continue

        LOADED_LANG.append(lang)


# insert rst block with correct image
def autodoc_process_docstring(app, what, name, obj, options, lines):
    def get_block_png(lang_, opcode_):
        if "readthedocs" in os.getcwd():
            return "/".join(["source", PATH_BLOCK_IMAGES, lang_, f"{lang_}_{opcode_}.png"])
        else:
            # use this path for local build
            return "/".join([PATH_BLOCK_IMAGES, lang_, f"{lang_}_{opcode_}.png"])

    try:
        # only insert blocks for methods
        if what == "method":
            lang = name.split(".")[1]
            if len(lang) == 2:
                download_pngs(lang)

                # get opcode from wrapped function
                opcode = inspect.getsourcelines(obj)[0][-1].strip(" ").split("(")[0].split(".")[-1]
                # opcode = opcode.replace("pystage_", "")
                if "pystage_" in opcode:
                    return lines
                path = get_block_png(lang, opcode)

                # insert rst figure block, care to put in empty lines above and below.
                for i in range(3):
                    lines.insert(1, "")
                lines.insert(4, f".. figure:: {path}")
                lines.insert(5, "    :height: 50")
                for i in range(3):
                    lines.insert(6, "")

                # add warning if png not found. This may have several reasons.
                # if not os.path.exists("/".join(["source", path])) and \
                #         not os.path.exists("/".join(["source", "source", path])):
                #     lines.insert(7, r'Image not found. Maybe function is not yet implemented for this language. '
                #                     r'Also the naming of the block may not the same as the corresponding function.\n If '
                #                     r'this error keeps existing check if this function is really a scratch block.')

    except IndexError:
        return lines

    return lines


def setup(app):
    app.connect('autodoc-process-docstring', autodoc_process_docstring)

